import datetime
from logging import Logger

from decimal import Decimal

import colorama
from sqlalchemy.orm import Session, Query
from typing import Optional, List

from sto.identityprovider import IdentityProvider
from sto.models.tokenscan import _TokenScanStatus
from sto.time import friendly_time


class NeedsTokenScan(Exception):
    pass


class CapTableEntry:
    """Reveal real world identity behind an address.

    Used when priting out cap table.
    """

    def __init__(self, name: str, address: str, balance: Decimal, updated_at: datetime.datetime):
        """
        :param name: Entity name or person name
        :param address: Crypto address where STOs are delivered
        """
        self.name = name
        self.address = address
        self.balance = balance
        self.percent = None
        self.updated_at = updated_at


class CapTableInfo:
    """Information about the cap table.

    For command line and web UIs to use.
    """

    def __init__(self, token_status: _TokenScanStatus, last_token_transfer_at, total_balance, entries):
        self.token_status = token_status
        self.last_token_transfer_at = last_token_transfer_at
        self.entries = entries
        self.total_balance = total_balance


def sort_entries(entries: List[CapTableEntry], order_by: str, order_direction: str):
    """Sort constructed cap table in place.

    Match friendly sort order to its underlying SQLite query.
    """
    if order_by == "balance":
        key = lambda entry: entry.balance
    elif order_by == "name":
        key = lambda entry: entry.name
    elif order_by == "updated":
        key = lambda entry: entry.updated_at
    elif order_by == "address":
        key = lambda entry: entry.address
    else:
        raise TypeError("Unknown sort order")

    if order_direction == "asc":
        entries.sort(key=key)
    elif order_direction == "desc":
        entries.sort(key=key, reverse=True)
    else:
        raise TypeError("Unknown sort direction")


def generate_cap_table(logger: Logger,
              dbsession: Session,
              token_address: str,
              order_by: str,
              order_direction: str,
              identity_provider: IdentityProvider,
              include_empty: bool,
              TokenScanStatus: type,
              TokenHolderAccount: type,
              no_name="<Unknown>") -> CapTableInfo:
    """Print out cap table.

    :param sort_order: "balance", "name", "updated", "address"
    :param include_empty: Include accounts that hold balance in the past
    :param TokenScanStatus: Token scan model used
    :param TokenHolderAccount: Token balance model used
    :return: List of CapTable entries
    """

    status = dbsession.query(TokenScanStatus).filter_by(address=token_address).one_or_none()  # type: TokenScanStatus
    if not status or status.end_block is None:
        raise NeedsTokenScan("No token {} balances available in the local database. Please run sto token-scan first.".format(token_address))

    q = status.get_accounts(include_empty)

    results = []
    total_balance = Decimal(0)
    last_token_transfer_at = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    for holder in q:

        id_check = identity_provider.get_identity(holder.address)
        if id_check:
            name = id_check.name
        else:
            name = no_name

        decimal_balance = holder.get_decimal_balance()

        entry = CapTableEntry(name, holder.address, decimal_balance, holder.last_block_updated_at)

        if entry.updated_at > last_token_transfer_at:
            last_token_transfer_at = entry.updated_at
        results.append(entry)

        if  decimal_balance > 0:  # Ignore cases where we cannot detect mint transaction
            total_balance += decimal_balance

    sort_entries(results, order_by, order_direction)

    # Retrofit decimal balances after we know the total sum
    if total_balance > 0:
        for r in results:
            r.percent = r.balance / total_balance

    info = CapTableInfo(status, last_token_transfer_at, total_balance, results)

    return info


def print_cap_table(info: CapTableInfo, max_entries: int, accuracy: int):
    """Console cap table printer"""

    if not info.token_status.end_block_timestamp:
        print("{}Token address {} not scanned. Please run sto token-scan first.{}".format(colorama.Fore.RED, info.token_status.address, colorama.Fore.RESET))
        return

    print("Token address: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, info.token_status.address, colorama.Fore.RESET))
    print("Name: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, info.token_status.name, colorama.Fore.RESET))
    print("Symbol: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, info.token_status.symbol, colorama.Fore.RESET))
    print("Total supply: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, info.token_status.total_supply, colorama.Fore.RESET))
    print("Accounted supply: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, info.total_balance, colorama.Fore.RESET))
    print("Holder count: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, len(info.entries), colorama.Fore.RESET))
    print("Cap table database updated at: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, friendly_time(info.token_status.end_block_timestamp), colorama.Fore.RESET))
    print("Last token transfer at at: {}{}{}".format(colorama.Fore.LIGHTCYAN_EX, friendly_time(info.last_token_transfer_at), colorama.Fore.RESET))

    print_entries = info.entries[0:max_entries]

    table = []

    balance_q = Decimal(10) ** Decimal(-accuracy)
    percent_q = Decimal("0.01")

    # Tuplify
    for idx, entry in enumerate(print_entries, start=1):
        table.append((
            idx,
            entry.name,
            entry.address,
            entry.updated_at,
            "{:,}".format(entry.balance.quantize(balance_q)),
            str(((entry.percent or 0) * Decimal(100)).quantize(percent_q)),
        ))

    from tabulate import tabulate  # https://bitbucket.org/astanin/python-tabulate
    output = tabulate(table, headers=["#", "Name", "Address", "Last transfer", "Balance", "%"], disable_numparse=True)
    print(output)









"""Distribute tokenised shares on Ethereum."""
from decimal import Decimal
from logging import Logger

import colorama
from tqdm import tqdm

from sto.distribution import DistributionEntry
from sto.ethereum.txservice import EthereumStoredTXService

from sto.ethereum.utils import get_abi, check_good_private_key, create_web3
from sto.ethereum.exceptions import BadContractException
from sto.models.broadcastaccount import _PreparedTransaction

from sto.models.implementation import BroadcastAccount, PreparedTransaction
from sqlalchemy.orm import Session
from typing import Union, Optional, List, Tuple
from web3 import Web3
from web3.exceptions import BadFunctionCallOutput


class NotEnoughTokens(Exception):
    pass



def distribute_tokens(logger: Logger,
                          dbsession: Session,
                          network: str,
                          ethereum_node_url: Union[str, Web3],
                          ethereum_abi_file: Optional[str],
                          ethereum_private_key: Optional[str],
                          ethereum_gas_limit: Optional[int],
                          ethereum_gas_price: Optional[int],
                          token_address: str,
                          dists: List[DistributionEntry]) -> Tuple[int, int]:
    """Sends tokens to their first owners in primary markets."""

    check_good_private_key(ethereum_private_key)

    abi = get_abi(ethereum_abi_file)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    logger.info("Starting creating distribution transactions for %s token from nonce %s", token_address, service.get_next_nonce())

    total = sum([dist.amount * 10**18 for dist in dists])

    available = service.get_raw_token_balance(token_address, abi)
    if total > available:
        raise NotEnoughTokens("Not enough tokens for distribution. Account {} has {} raw token balance, needed {}".format(service.get_or_create_broadcast_account().address, available, total))

    new_distributes = old_distributes = 0

    for d in tqdm(dists):
        if not service.is_distributed(d.external_id, token_address):
            # Going to tx queue
            raw_amount = int(d.amount * 10**18)
            note = "Distributing tokens, raw amount: {}".format(raw_amount)
            service.distribute_tokens(d.external_id, d.address, raw_amount, token_address, abi, note)
            new_distributes += 1
        else:
            # CSV reimports
            old_distributes += 1

    logger.info("Prepared transactions for broadcasting for network %s", network)
    return new_distributes, old_distributes



def distribute_single(logger: Logger,
                          dbsession: Session,
                          network: str,
                          ethereum_node_url: Union[str, Web3],
                          ethereum_abi_file: Optional[str],
                          ethereum_private_key: Optional[str],
                          ethereum_gas_limit: Optional[int],
                          ethereum_gas_price: Optional[int],
                          token_address: str,
                          ext_id: str,
                          email: str,
                          name: str,
                          to_address: str,
                          amount: Decimal) -> bool:
    """Send out a single transfer.

    :return: True if a new tx for broadcasting was created
    """

    d = DistributionEntry(ext_id, email, name, to_address, amount)

    check_good_private_key(ethereum_private_key)

    abi = get_abi(ethereum_abi_file)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    logger.info("Starting creating distribution transactions for %s token from nonce %s", token_address, service.get_next_nonce())

    total = d.amount * 10**18

    available = service.get_raw_token_balance(token_address, abi)
    if total > available:
        raise NotEnoughTokens("Not enough tokens for distribution. Account {} has {} raw token balance, needed {}".format(service.get_or_create_broadcast_account().address, available, total))

    if not service.is_distributed(d.external_id, token_address):
        # Going to tx queue
        raw_amount = int(d.amount * 10**18)
        note = "Distributing tokens, raw amount: {}".format(raw_amount)
        service.distribute_tokens(d.external_id, d.address, raw_amount, token_address, abi, note)
        logger.info("New broadcast has been created")
        return True
    else:
        logger.error("Already distributed")
        return False




import json
import os

from typing import Optional

import rlp
from eth_abi import encode_abi
from web3 import Web3, HTTPProvider
from web3.contract import Contract
from web3.utils.abi import get_constructor_abi, merge_args_and_kwargs
from web3.utils.events import get_event_data
from web3.utils.filters import construct_event_filter_params
from eth_utils import keccak, to_checksum_address, to_bytes, is_hex_address, is_checksum_address
from web3.utils.contracts import encode_abi

class NoNodeConfigured(Exception):
    pass


class NeedPrivateKey(Exception):
    pass



def check_good_node_url(node_url: str):
    if not node_url:
        raise NoNodeConfigured("You need to give --ethereum-node-url command line option or set it up in a config file")


def check_good_private_key(private_key_hex: str):
    if not private_key_hex:
        raise NeedPrivateKey("You need to give --ethereum-private-key command line option or set it up in a config file")


def get_abi(abi_file: Optional[str]):

    if not abi_file:
        # Use built-in solc output drop
        abi_file = os.path.join(os.path.dirname(__file__), "contracts-flattened.json")

    with open(abi_file, "rt") as inp:
        return json.load(inp)


def create_web3(url: str) -> Web3:
    """Web3 initializer."""

    if isinstance(url, Web3):
        # Shortcut for testing
        return url
    else:
        return Web3(HTTPProvider(url))



def mk_contract_address(sender: str, nonce: int) -> str:
    """Create a contract address using eth-utils.

    https://ethereum.stackexchange.com/a/761/620
    """
    sender_bytes = to_bytes(hexstr=sender)
    raw = rlp.encode([sender_bytes, nonce])
    h = keccak(raw)
    address_bytes = h[12:]
    return to_checksum_address(address_bytes)


# Sanity check
assert mk_contract_address(to_checksum_address("0x6ac7ea33f8831ea9dcc53393aaa88b25a785dbf0"), 1) == to_checksum_address("0x343c43a37d37dff08ae8c4a11544c718abb4fcf8")


def validate_ethereum_address(address: str):
    """Clever Ethereum address validator.

    Assume all lowercase addresses are not checksummed.
    """

    if len(address) < 42:
        raise ValueError("Not an Ethereum address: {}".format(address))

    try:
        if not is_hex_address(address):
            raise ValueError("Not an Ethereum address: {}".format(address))
    except UnicodeEncodeError:
        raise ValueError("Could not decode: {}".format(address))

    # Check if checksummed address if any of the letters is upper case
    if any([c.isupper() for c in address]):
        if not is_checksum_address(address):
            raise ValueError("Not a checksummed Ethereum address: {}".format(address))


def get_constructor_arguments(contract: Contract, args: Optional[list]=None, kwargs: Optional[dict]=None):
    """Get constructor arguments for Etherscan verify.

    https://etherscanio.freshdesk.com/support/solutions/articles/16000053599-contract-verification-constructor-arguments
    """

    # return contract._encode_constructor_data(args=args, kwargs=kwargs)

    constructor_abi = get_constructor_abi(contract.abi)

    if args is not None:
        return contract._encode_abi(constructor_abi, args)[2:]  # No 0x
    else:
        constructor_abi = get_constructor_abi(contract.abi)
        kwargs = kwargs or {}
        arguments = merge_args_and_kwargs(constructor_abi, [], kwargs)
        # deploy_data = add_0x_prefix(
        #    contract._encode_abi(constructor_abi, arguments)
        #)

        # TODO: Looks like recent Web3.py ABI change
        deploy_data = encode_abi(contract.web3, constructor_abi, arguments)
        return deploy_data


def getLogs(self,
    argument_filters=None,
    fromBlock=None,
    toBlock="latest",
    address=None,
    topics=None):
    """Get events using eth_getLogs API.

    This is a stateless method, as opposite to createFilter.
    It can be safely called against nodes which do not provide eth_newFilter API, like Infura.

    :param argument_filters:
    :param fromBlock:
    :param toBlock:
    :param address:
    :param topics:
    :return:
    """

    if fromBlock is None:
        raise TypeError("Missing mandatory keyword argument to getLogs: fromBlock")

    abi = self._get_event_abi()

    argument_filters = dict()

    _filters = dict(**argument_filters)

    # Construct JSON-RPC raw filter presentation based on human readable Python descriptions
    # Namely, convert event names to their keccak signatures
    data_filter_set, event_filter_params = construct_event_filter_params(
        abi,
        contract_address=self.address,
        argument_filters=_filters,
        fromBlock=fromBlock,
        toBlock=toBlock,
        address=address,
        topics=topics,
    )

    # Call JSON-RPC API
    logs = self.web3.eth.getLogs(event_filter_params)

    # Convert raw binary data to Python proxy objects as described by ABI
    for entry in logs:
        yield get_event_data(abi, entry)







import binascii
from logging import Logger

import colorama
from eth_account.account import Account
from eth_utils import to_hex

CONFIG_FILE_TEMPLATE = """

#
# Your STO configuration file for {network}
#

network = {network}

# Where to connect for Parity or Geth JSON-RPC API
ethereum-node-url = http://localhost:8545

# The private key for your generated Ethereum account {address}
ethereum-private-key = {private_key}
"""


def create_account_console(logger: Logger, network: str):
    """Creates an Ethereum account and echoes the details to console."""

    acc = Account.create()
    private_key = to_hex(acc.privateKey)
    logger.info("Account address: %s", acc.address)
    logger.info("Account private key: %s", private_key)

    config = CONFIG_FILE_TEMPLATE.format(private_key=private_key, network=network, address=acc.address)
    config_file_name = "myconfig.ini"

    print()
    print("Create a file {}{}{} and paste in the following content: {}{}{}".format(colorama.Fore.LIGHTBLUE_EX, config_file_name, colorama.Fore.RESET, colorama.Fore.LIGHTBLACK_EX, config, colorama.Fore.RESET))
    print()
    print("After this you can run {}sto --config-file={} diagnose{}".format(colorama.Fore.LIGHTBLUE_EX, config_file_name, colorama.Fore.RESET))


"""Issuing out tokenised shares."""
from decimal import Decimal
from logging import Logger

import colorama
import requests
from tqdm import tqdm

from sto.ethereum.txservice import EthereumStoredTXService, verify_on_etherscan

from sto.ethereum.utils import get_abi, check_good_private_key, create_web3
from sto.ethereum.exceptions import BadContractException

from sto.models.implementation import BroadcastAccount, PreparedTransaction
from sqlalchemy.orm import Session
from typing import Union, Optional, List, Iterable
from web3 import Web3
from web3.exceptions import BadFunctionCallOutput


class NeedAPIKey(RuntimeError):
    pass


def deploy_token_contracts(logger: Logger,
                          dbsession: Session,
                          network: str,
                          ethereum_node_url: Union[str, Web3],
                          ethereum_abi_file: Optional[str],
                          ethereum_private_key: Optional[str],
                          ethereum_gas_limit: Optional[int],
                          ethereum_gas_price: Optional[int],
                          name: str,
                          symbol: str,
                          amount: int,
                          transfer_restriction: str):
    """Issue out a new Ethereum token."""

    assert type(amount) == int
    decimals = 18  # Everything else is bad idea

    check_good_private_key(ethereum_private_key)

    abi = get_abi(ethereum_abi_file)

    web3 = create_web3(ethereum_node_url)

    # We do not have anything else implemented yet
    assert transfer_restriction == "unrestricted"

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    # Deploy security token
    note = "Deploying token contract for {}".format(name)
    deploy_tx1 = service.deploy_contract("SecurityToken", abi, note, constructor_args={"_name": name, "_symbol": symbol})  # See SecurityToken.sol

    # Deploy transfer agent
    note = "Deploying unrestricted transfer policy for {}".format(name)
    deploy_tx2 = service.deploy_contract("UnrestrictedTransferAgent", abi, note)

    # Set transfer agent
    note = "Whitelisting deployment account for {} issuer control".format(name)
    contract_address = deploy_tx1.contract_address
    update_tx1 = service.interact_with_contract("SecurityToken", abi, contract_address, note, "addAddressToWhitelist", {"addr": deploy_tx1.get_from()})
    assert update_tx1.contract_deployment == False

    # Set transfer agent
    note = "Making transfer restriction policy for {} effective".format(name)
    contract_address = deploy_tx1.contract_address
    update_tx2 = service.interact_with_contract("SecurityToken", abi, contract_address, note, "setTransactionVerifier", {"newVerifier": deploy_tx2.contract_address})

    # Issue out initial shares
    note = "Creating {} initial shares for {}".format(amount, name)
    contract_address = deploy_tx1.contract_address
    amount_18 = int(amount * 10**decimals)
    update_tx3 = service.interact_with_contract("SecurityToken", abi, contract_address, note, "issueTokens", {"value": amount_18})

    logger.info("Prepared transactions for broadcasting for network %s", network)
    logger.info("STO token contract address will be %s%s%s", colorama.Fore.LIGHTGREEN_EX, deploy_tx1.contract_address, colorama.Fore.RESET)
    return [deploy_tx1, deploy_tx2, update_tx1, update_tx2, update_tx3]


def contract_status(logger: Logger,
                          dbsession: Session,
                          network: str,
                          ethereum_node_url: str,
                          ethereum_abi_file: str,
                          ethereum_private_key: str,
                          ethereum_gas_limit: str,
                          ethereum_gas_price: str,
                          token_contract: str):
    """Poll STO contract status."""

    abi = get_abi(ethereum_abi_file)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)
    contract = service.get_contract_proxy("SecurityToken", abi, token_contract)

    try:
        logger.info("Name: %s", contract.functions.name().call())
        logger.info("Symbol: %s", contract.functions.symbol().call())
        supply = contract.functions.totalSupply().call()
        human_supply = Decimal(supply) / Decimal(10 ** contract.functions.decimals().call())
        raw_balance = contract.functions.balanceOf(service.get_or_create_broadcast_account().address).call()
        normal_balance = Decimal(raw_balance) / Decimal(10 ** contract.functions.decimals().call())
        logger.info("Total supply: %s", human_supply)
        logger.info("Decimals: %d", contract.functions.decimals().call())
        logger.info("Owner: %s", contract.functions.owner().call())
        logger.info("Broadcast account token balance: %f", normal_balance)
        logger.info("Transfer verified: %s", contract.functions.transferVerifier().call())
    except BadFunctionCallOutput as e:
        raise BadContractException("Looks like this is not a token contract address. Please check on EtherScan that the address presents the token contract")

    return {
        "name": contract.functions.name().call(),
        "symbol": contract.functions.symbol().call(),
        "totalSupply": contract.functions.totalSupply().call(),
        "broadcastBalance": raw_balance,
    }



def verify_source_code(logger: Logger,
              dbsession: Session,
              network: str,
              etherscan_api_key: str,
):
    """Verify source code of all unverified deployment transactions."""

    if not etherscan_api_key:
        raise NeedAPIKey("You need to give EtherScan API key in the configuration file. Get one from https://etherscan.io")

    unverified_txs = dbsession.query(PreparedTransaction).filter_by(verified_at=None, result_transaction_success=True, contract_deployment=True)

    logger.info("Found %d unverified contract deployments on %s", unverified_txs.count(), network)

    if unverified_txs.count() == 0:
        logger.info("No transactions to verify.")
        return []

    unverified_txs = list(unverified_txs)

    # HTTP keep-alive
    session = requests.Session()

    # https://stackoverflow.com/questions/41985993/tqdm-show-progress-for-a-generator-i-know-the-length-of
    for tx in unverified_txs:  # type: _PreparedTx
        logger.info("Verifying %s for %s", tx.contract_address, tx.human_readable_description)
        verify_on_etherscan(logger, network, tx, etherscan_api_key, session)
        dbsession.commit()  # Try to minimise file system sync issues

    return unverified_txs


def past_issuances(logger: Logger, dbsession: Session) -> Iterable[PreparedTransaction]:
    """Get list of past issuances."""

    txs = dbsession.query(PreparedTransaction).filter_by(contract_deployment=True)
    for tx in txs:
        if tx.is_token_contract_deployment():
            yield tx




from logging import Logger

import requests
import time
from eth_account.internal.transactions import assert_valid_fields
from sto.ethereum.utils import mk_contract_address, get_constructor_arguments
from sto.models.broadcastaccount import _BroadcastAccount, _PreparedTransaction
from sto.models.utils import now
from eth_account import Account
from eth_utils import to_checksum_address

from sqlalchemy.orm import Session, Query
from typing import Optional, Iterable
from web3 import Web3
from web3.contract import Contract

from sto.models.implementation import _BroadcastAccount


class NetworkAndDatabaseNonceOutOfSync(Exception):
    pass



class AddressConfigurationMismatch(Exception):
    pass


class CouldNotVerifyOnEtherScan(Exception):
    pass


class EthereumStoredTXService:
    """A transaction service that writes entries to a local database before trying to broadcast them to the blockchain."""

    #: Can't trust auto estimate
    SPECIAL_GAS_LIMIT_FOR_CONTRACT_DEPLOYMENT = 3141619  # Number from Ethereum tester, cannot exceed this

    #: Can't trust auto estimate
    SPECIAL_GAS_LIMIT_FOR_NORMAL_TX = 350333

    def __init__(self, network: str, dbsession: Session, web3: Web3, private_key_hex: str, gas_price, gas_limit, broadcast_account_model, prepared_tx_model):

        assert isinstance(web3, Web3)

        self.network = network  # "kovan"
        self.dbsession = dbsession
        self.web3 = web3
        self.private_key_hex = private_key_hex
        self.account = Account.privateKeyToAccount(private_key_hex)

        # SQLAlchemy models, allow caller to supply their own
        self.broadcast_account_model = broadcast_account_model
        self.prepared_tx_model = prepared_tx_model

        self.gas_price = gas_price


        if gas_limit:
            assert type(gas_limit) == int

        self.gas_limit = gas_limit

    @property
    def address(self):
        return self.account.address

    def get_or_create_broadcast_account(self):
        """
        :return:
        """

        # Some early prototype sanity checks
        assert self.address.startswith("0x")
        assert self.network in ("kovan", "ethereum", "testing", "ropsten")  # TODO: Sanity check - might want to remove this

        account = self.dbsession.query(self.broadcast_account_model).filter_by(network=self.network, address=self.address).one_or_none()
        if not account:
            account = self.broadcast_account_model(network=self.network, address=self.address)
            self.dbsession.add(account)
            self.dbsession.flush()

        return account

    def get_next_nonce(self):
        broadcast_account = self.get_or_create_broadcast_account()
        return broadcast_account.current_nonce

    def ensure_accounts_in_sync(self):
        """Make sure that our internal nonce and external nonce looks correct."""

        broadcast_account = self.get_or_create_broadcast_account()

        tx_count = self.web3.eth.getTransactionCount(broadcast_account.address)

        if tx_count != broadcast_account.current_nonce:
            NetworkAndDatabaseNonceOutOfSync("Nonced out of sync. Network: {}, database: {}. Maybe you have a pending broadcasts propagating?".format(tx_count, broadcast_account.current_nonce))

    def allocate_transaction(self,
                             broadcast_account: _BroadcastAccount,
                             receiver: Optional[str],
                             contract_address: Optional[str],
                             contract_deployment: bool,
                             nonce: int,
                             note: str,
                             unsigned_payload: dict,
                             gas_price: Optional[int],
                             gas_limit: Optional[int],
                             external_id: Optional[str]=None,
                             ) -> _PreparedTransaction:
        """Put a transaction to the pending queue of the current broadcast list."""

        if receiver:
            assert receiver.startswith("0x")

        assert contract_deployment in (True, False)

        assert broadcast_account.current_nonce == nonce

        assert type(unsigned_payload) == dict

        assert_valid_fields(unsigned_payload)

        tx = self.prepared_tx_model()  # type: _PreparedTransaction
        tx.nonce = nonce
        tx.human_readable_description = note
        tx.receiver = receiver
        tx.contract_address = contract_address
        tx.contract_deployment = contract_deployment
        tx.unsigned_payload = unsigned_payload
        tx.external_id = external_id
        broadcast_account.txs.append(tx)
        broadcast_account.current_nonce += 1
        return tx

    def generate_tx_data(self, nonce: int, contract_tx=False) -> dict:
        """Generate transaction control parameters.

        :param contract: We use a special hardcoded gas estimate for 4,000,000 when deploying contracts. Kovan misestimates the cost of deploying SecurityToken and thus the transaction always fails with the auto estimate.
        """

        # See TRANSACTION_VALID_VALUES
        tx_data = {}
        tx_data["nonce"] = nonce


        # Mikko's rule of thumb estimator because local accounts do not estimate gas too well
        if self.gas_limit:
            tx_data["gas"] = self.gas_limit
        elif contract_tx:
            tx_data["gas"] = EthereumStoredTXService.SPECIAL_GAS_LIMIT_FOR_CONTRACT_DEPLOYMENT
        else:
            tx_data["gas"] = EthereumStoredTXService.SPECIAL_GAS_LIMIT_FOR_NORMAL_TX

        if self.gas_price:
            tx_data["gasPrice"] = self.gas_price

        return tx_data

    def deploy_contract(self, contract_name: str, abi: dict, note: str, constructor_args=None) -> _PreparedTransaction:
        """Deploys a contract."""

        if not constructor_args:
            constructor_args = {}

        abi_data = abi[contract_name]

        assert "source" in abi_data, "We need to have special postprocessed ABI data bundle, as we need the contract source code for EtherScan verification"

        contract_class = Contract.factory(
            web3=self.web3,
            abi=abi_data["abi"],
            bytecode=abi_data["bytecode"],
            bytecode_runtime=abi_data["bytecode_runtime"],
            )

        broadcast_account = self.get_or_create_broadcast_account()

        next_nonce = self.get_next_nonce()

        # Creates a dict for signing
        tx_data = self.generate_tx_data(next_nonce, contract_tx=True)
        constructed_txn = contract_class.constructor(**constructor_args).buildTransaction(tx_data)

        constructor_arguments = get_constructor_arguments(contract_class, kwargs=constructor_args)

        derived_contract_address = mk_contract_address(self.address, next_nonce)
        derived_contract_address = to_checksum_address(derived_contract_address.lower())

        constructed_txn["to"] = "" # Otherwise database serializer complains about bytes string

        tx = self.allocate_transaction(
            broadcast_account=broadcast_account,
            receiver=None,
            contract_address=derived_contract_address,
            contract_deployment=True,
            nonce=next_nonce,
            note=note,
            unsigned_payload=constructed_txn,
            gas_price=self.gas_price,
            gas_limit=self.gas_limit,
        )

        self.dbsession.flush()  # Populate other_data

        tx.abi = abi_data
        tx.constructor_arguments = constructor_arguments
        assert tx.compiler_version
        assert tx.flattened_source_code

        return tx

    def get_contract_proxy(self, contract_name: str, abi: dict, address: str) -> Contract:
        """Get web3.Contract to interact directly with the network"""
        abi_data = abi[contract_name]

        contract_class = Contract.factory(
            web3=self.web3,
            abi=abi_data["abi"],
            bytecode=abi_data["bytecode"],
            bytecode_runtime=abi_data["bytecode_runtime"],
            )

        return contract_class(address=to_checksum_address(address))

    def interact_with_contract(self, contract_name: str, abi: dict, address: str, note: str, func_name: str, args=None, receiver=None) -> _PreparedTransaction:
        """Does a transaction against a contract."""

        assert address.startswith("0x")

        if not args:
            args = {}

        contract = self.get_contract_proxy(contract_name, abi, address)
        broadcast_account = self.get_or_create_broadcast_account()

        next_nonce = self.get_next_nonce()

        func = getattr(contract.functions, func_name)

        tx_data = self.generate_tx_data(next_nonce)
        constructed_txn = func(**args).buildTransaction(tx_data)

        tx = self.allocate_transaction(
            broadcast_account=broadcast_account,
            receiver=receiver,
            contract_address=address,
            contract_deployment=False,
            nonce=next_nonce,
            note=note,
            unsigned_payload=constructed_txn,
            gas_price=self.gas_price,
            gas_limit=self.gas_limit,
        )
        self.dbsession.flush()
        return tx

    def is_distributed(self, external_id: str, contract_address: str) -> bool:
        """Prevent us sending the same transaction twice."""
        if self.dbsession.query(self.prepared_tx_model).filter_by(external_id=external_id, contract_address=contract_address).one_or_none():
            return True
        return False

    def distribute_tokens(self, external_id: str, receiver_address: str, raw_amount: int, token_address: str, abi: dict, note: str, contract_name="ERC20", func_name="transfer", receiver=None) -> _PreparedTransaction:
        """Send out tokens."""

        assert receiver_address.startswith("0x")
        assert token_address.startswith("0x")
        assert type(raw_amount) == int
        assert raw_amount >= 1

        # Prevent us sending the same transaction twice
        if self.dbsession.query(self.prepared_tx_model).filter_by(external_id=external_id, contract_address=token_address).one_or_none():
            raise RuntimeError("Already distributed token:{} id:{}".format(token_address, external_id))

        contract = self.get_contract_proxy(contract_name, abi, token_address)
        broadcast_account = self.get_or_create_broadcast_account()

        next_nonce = self.get_next_nonce()

        args = [receiver_address, raw_amount]
        func = getattr(contract.functions, func_name)

        tx_data = self.generate_tx_data(next_nonce)
        constructed_txn = func(*args).buildTransaction(tx_data)

        tx = self.allocate_transaction(
            broadcast_account=broadcast_account,
            receiver=receiver_address,
            contract_address=token_address,
            contract_deployment=False,
            nonce=next_nonce,
            note=note,
            unsigned_payload=constructed_txn,
            gas_price=self.gas_price,
            gas_limit=self.gas_limit,
            external_id=external_id,
        )
        self.dbsession.flush()
        return tx

    def get_raw_token_balance(self, token_address: str, abi: dict, contract_name="ERC20Basic", func_name="balanceOf") -> int:
        """Check that we have enough token balance for distribute operations."""

        assert token_address.startswith("0x")

        contract = self.get_contract_proxy(contract_name, abi, token_address)
        broadcast_account = self.get_or_create_broadcast_account()

        args = {
            "who": broadcast_account.address,
        }
        func = getattr(contract.functions, func_name)

        result = func(**args).call()
        return result

    def get_pending_broadcasts(self) -> Query:
        """All transactions that need to be broadcasted."""
        return self.dbsession.query(self.prepared_tx_model).filter_by(broadcasted_at=None).order_by(self.prepared_tx_model.nonce).join(self.broadcast_account_model).filter_by(network=self.network)

    def get_unmined_txs(self) -> Query:
        """All transactions that do not yet have a block assigned."""
        return self.dbsession.query(self.prepared_tx_model).filter(self.prepared_tx_model.txid != None).filter_by(result_block_num=None).join(self.broadcast_account_model).filter_by(network=self.network)

    def get_last_transactions(self, limit: int) -> Query:
        """Fetch latest transactions."""
        assert type(limit) == int
        return self.dbsession.query(self.prepared_tx_model).order_by(self.prepared_tx_model.created_at.desc()).limit(limit)

    def broadcast(self, tx: _PreparedTransaction):
        """Push transactions to Ethereum network."""

        if tx.broadcast_account.address != self.address:
            raise AddressConfigurationMismatch("Could not broadcast due to address mismatch. A pendign transaction was created for account {}, but we are using configured account {}".format(tx.broadcast_account.addres, self.address))

        tx_data = tx.unsigned_payload
        signed = self.web3.eth.account.signTransaction(tx_data, self.private_key_hex)
        tx.txid = signed.hash.hex()
        self.web3.eth.sendRawTransaction(signed.rawTransaction)
        tx.broadcasted_at = now()
        return tx

    def update_status(self, tx: _PreparedTransaction):
        """Update tx status from Etheruem network."""

        assert tx.txid

        # https://web3py.readthedocs.io/en/stable/web3.eth.html#web3.eth.Eth.getTransactionReceipt
        receipt = self.web3.eth.getTransactionReceipt(tx.txid)
        if receipt:
            tx.result_block_num = receipt["blockNumber"]

            # https://ethereum.stackexchange.com/a/6003/620
            if receipt["status"] == 0:
                tx.result_transaction_success = False
                tx.result_transaction_reason = "Transaction failed"  # TODO: Need some logic to separate failure modes
            else:
                tx.result_transaction_success = True

        tx.result_fetched_at = now()
        return tx

    @classmethod
    def print_transactions(self, txs: Iterable[_PreparedTransaction]):
        """Print transaction status to the console"""

        from tabulate import tabulate # https://bitbucket.org/astanin/python-tabulate
        import colorama # https://pypi.org/project/colorama/

        colorama.init()

        table = []
        for tx in txs:

            status = tx.get_status()
            if status == "waiting":
                status = colorama.Fore.BLUE + status + colorama.Fore.RESET
            elif status == "broadcasted":
                status = colorama.Fore.YELLOW + status + colorama.Fore.RESET
            elif status == "mining":
                status = colorama.Fore.YELLOW + status + colorama.Fore.RESET
            elif status in ("success", "verified"):
                status = colorama.Fore.GREEN + status + colorama.Fore.RESET
                status += ":" + str(tx.result_block_num)
            elif status == "failed":
                status = colorama.Fore.RED + status + colorama.Fore.RESET
                status += ":" + str(tx.result_block_num)
            else:
                raise RuntimeError("Does not compute")

            table.append((tx.txid, status, tx.nonce, tx.get_from(), tx.get_to(), tx.human_readable_description[0:64]))

        print(tabulate(table, headers=["TXID", "Status and block", "Nonce", "From", "To", "Note"]))


def verify_on_etherscan(logger: Logger, network: str, tx: _PreparedTransaction, api_key: str, session, timeout=120):
    """Verify a contrcact deployment on Etherscan.

    Uses https://etherscan.io/apis#contracts
    """

    assert network in ("ethereum", "kovan")
    if network == "kovan":
        url = "https://api-kovan.etherscan.io/api"
    else:
        url = "https://api.etherscan.io/api"

    assert tx.result_transaction_success
    assert tx.contract_deployment

    source = tx.flattened_source_code
    compiler = tx.compiler_version
    address = tx.contract_address
    constructor_arguments = tx.constructor_arguments
    contract_name = tx.contract_name

    data = {
        "apikey": api_key,
        "module": "contract",
        "contractaddress": address,
        "action": "verifysourcecode",
        "sourceCode": source,
        "contractname": contract_name,
        "compilerversion": "v" + compiler,  # https://etherscan.io/solcversions
        "constructorArguements": constructor_arguments[2:],  # Remove leading 0x
        "optimizationUsed": 1,  # TODO: Hardcoded
        "runs": 500, # TODO: Hardcoded
    }

    info_data = data.copy()
    del info_data["sourceCode"]  # Too verbose
    del info_data["apikey"]  # Security
    logger.info("Calling EtherScan API as: %s", info_data)

    #
    # Step 1: EtherScan validates input and gives us a ticket id to track the submission status
    #

    resp = session.post(url, data)
    # {'status': '0', 'message': 'NOTOK', 'result': 'Error!'}
    data = resp.json()

    logger.info("Etherscan replied %s", data)
    if data["status"] == "0":

        if "already verified" in data["result"]:
            return

        raise CouldNotVerifyOnEtherScan("Could not verify contract: " + address + " " + str(data))


    ticket = data["result"]

    #
    # Step 2: Poll for results
    #
    ready = False
    started = time.time()
    while not ready and time.time() < started + timeout:
        data = {
            "apikey": api_key,
            "module": "contract",
            "action": "checkverifystatus",
            "guid": ticket,
        }
        logger.info("Checking verification status on EtherScan API as: %s", info_data)
        resp = session.post(url, data)
        # {'status': '0', 'message': 'NOTOK', 'result': 'Error!'}
        data = resp.json()
        logger.info("Got reply %s", data)

        if data["result"] == "Pending in queue":
            # Keep polling
            #  {'status': '0', 'message': 'NOTOK', 'result': 'Pending in queue'}
            time.sleep(5)
            continue
        elif data["status"] == "0":
            # Produced binary did not match
            raise CouldNotVerifyOnEtherScan("Could not verify contract: " + address + " " + str(data))
        else:
            # All good
            assert data["status"] == "1"  # {'status': '1', 'message': 'OK', 'result': 'Pass - Verified'}
            break

    # Write to the database that we managed verify the contract
    tx.verified_at = now()
    tx.verification_info = data

    # //Submit Source Code for Verification
    # $.ajax({
    #     type: "POST",                       //Only POST supported
    #     url: "//api-kovan.etherscan.io/api", //Set to the  correct API url for Other Networks
    #     data: {
    #         apikey: $('#apikey').val(),                     //A valid API-Key is required
    #         module: 'contract',                             //Do not change
    #         action: 'verifysourcecode',                     //Do not change
    #         contractaddress: $('#contractaddress').val(),   //Contract Address starts with 0x...
    #         sourceCode: $('#sourceCode').val(),             //Contract Source Code (Flattened if necessary)
    #         contractname: $('#contractname').val(),         //ContractName
    #         compilerversion: $('#compilerversion').val(),   // see http://etherscan.io/solcversions for list of support versions
    #         optimizationUsed: $('#optimizationUsed').val(), //0 = Optimization used, 1 = No Optimization
    #         runs: 200,                                      //set to 200 as default unless otherwise
    #         constructorArguements: $('#constructorArguements').val(),   //if applicable
    #         libraryname1: $('#libraryname1').val(),         //if applicable, a matching pair with libraryaddress1 required
    #         libraryaddress1: $('#libraryaddress1').val(),   //if applicable, a matching pair with libraryname1 required
    #         libraryname2: $('#libraryname2').val(),         //if applicable, matching pair required
    #         libraryaddress2: $('#libraryaddress2').val(),   //if applicable, matching pair required
    #         libraryname3: $('#libraryname3').val(),         //if applicable, matching pair required
    #         libraryaddress3: $('#libraryaddress3').val(),   //if applicable, matching pair required
    #         libraryname4: $('#libraryname4').val(),         //if applicable, matching pair required
    #         libraryaddress4: $('#libraryaddress4').val(),   //if applicable, matching pair required
    #         libraryname5: $('#libraryname5').val(),         //if applicable, matching pair required
    #         libraryaddress5: $('#libraryaddress5').val(),   //if applicable, matching pair required
    #         libraryname6: $('#libraryname6').val(),         //if applicable, matching pair required
    #         libraryaddress6: $('#libraryaddress6').val(),   //if applicable, matching pair required
    #         libraryname7: $('#libraryname7').val(),         //if applicable, matching pair required
    #         libraryaddress7: $('#libraryaddress7').val(),   //if applicable, matching pair required
    #         libraryname8: $('#libraryname8').val(),         //if applicable, matching pair required
    #         libraryaddress8: $('#libraryaddress8').val(),   //if applicable, matching pair required
    #         libraryname9: $('#libraryname9').val(),         //if applicable, matching pair required
    #         libraryaddress9: $('#libraryaddress9').val(),   //if applicable, matching pair required
    #         libraryname10: $('#libraryname10').val(),       //if applicable, matching pair required
    #         libraryaddress10: $('#libraryaddress10').val()  //if applicable, matching pair required
    #     },
    #     success: function (result) {
    #         console.log(result);
    #         if (result.status == "1") {
    #             //1 = submission success, use the guid returned (result.result) to check the status of your submission.
    #             // Average time of processing is 30-60 seconds
    #             document.getElementById("postresult").innerHTML = result.status + ";" + result.message + ";" + result.result;
    #             // result.result is the GUID receipt for the submission, you can use this guid for checking the verification status
    #         } else {
    #             //0 = error
    #             document.getElementById("postresult").innerHTML = result.status + ";" + result.message + ";" + result.result;
    #         }
    #         console.log("status : " + result.status);
    #         console.log("result : " + result.result);
    #     },
    #     error: function (result) {
    #         console.log("error!");
    #         document.getElementById("postresult").innerHTML = "Unexpected Error"
    #     }
    # });




"""Share distribution functionalities."""
import csv
from decimal import Decimal
from logging import Logger

from typing import List

from sto.ethereum.utils import validate_ethereum_address


class DistributionEntry:
    """Hold information about who is going to receive the share.

    Used to harmonise different input and output formats.
    """

    def __init__(self, external_id: str, email: str, name: str, address: str, amount: Decimal):
        """
        :param external_id: External transaction id - always unique - like a TXID for Ethereum or a fiat payment reference number.
        :param email: We pass email ids around as they are generally easy way to identity entities across systems in human readable manner - think Asian letters
        :param name: Entity name or person name
        :param address: Crypto address where STOs are delivered
        :param amount: Amount of STO
        """
        self.external_id = external_id
        self.email = email
        self.name = name
        self.address = address
        self.amount = amount

    def __json__(self):
        """Convert for JSON export."""
        return {
            "external_id": self.external_id,
            "name": self.name,
            "address": self.address,
            "email": self.email,
            "amount": str(self.amount)
        }


def read_csv(logger: Logger, fname) -> List[DistributionEntry]:
    """Read outgoing distribution records from CSV file.

    Columns are:
    * external_id
    * email
    * name
    * address
    * amount
    """

    logger.info("Reading CSV input %s", fname)
    with open(fname, "rt") as inp:
        reader = csv.DictReader(inp)
        rows = [row for row in reader]

    result = []

    # TODO: Address format is now hardcoded for Etheereum.
    for idx, row in enumerate(rows):
        addr = row["address"].strip()
        try:
            validate_ethereum_address(addr)
        except ValueError as e:
            raise ValueError("Invalid Ethereum address on row:" + str(idx+1) + " address:", addr +  " reason:" + str(e) + " external_id:" + row["external_id"]) from e

        amount = row["amount"].strip()
        try:
            amount = Decimal(amount)
        except ValueError as e:
            raise ValueError("Invalid decimal amounton row:" + str(idx+1) + " amount:", addr +  " reason:" + str(e) + " external_id:" + row["external_id"]) from e

        entry = DistributionEntry(row["external_id"], row["email"], row["name"], row["address"], amount)
        result.append(entry)

    return result






import datetime

import sqlalchemy as sa

from sqlalchemy import DateTime, processors
from sqlalchemy.dialects.sqlite import DATETIME as DATETIME_


def now() -> datetime.datetime:
    """Get the current time as timezone-aware UTC timestamp."""
    return datetime.datetime.now(datetime.timezone.utc)


class SQLITEDATETIME(DATETIME_):
    """Timezone aware datetime support for SQLite.

    This is implementation used for UTCDateTime.
    """

    @staticmethod
    def process(value):
        dt = processors.str_to_datetime(value)
        if dt:
            # Returns naive datetime, force it to UTC
            return dt.replace(tzinfo=datetime.timezone.utc)
        return dt

    def result_processor(self, dialect, coltype):
        return SQLITEDATETIME.process


class UTCDateTime(DateTime):
    """An SQLAlchemy DateTime column that explicitly uses timezone aware dates and only accepts UTC."""

    def __init__(self, *args, **kwargs):
        # If there is an explicit timezone we accept UTC only
        if "timezone" in kwargs:
            assert kwargs["timezone"] in (datetime.timezone.utc, True)

        kwargs = kwargs.copy()
        kwargs["timezone"] = True
        super(UTCDateTime, self).__init__(**kwargs)

    def _dialect_info(self, dialect):
        if dialect.name == "sqlite":
            # Becase SQLite does not support datetimes, we need to explicitly tell here to use our super duper DATETIME() hack subclass that hacks in timezone
            return {"impl": SQLITEDATETIME()}
        else:
            return super(UTCDateTime, self)._dialect_info(dialect)


class TimeStampedBaseModel:
    """
    Base model with UUID as PK and have create & update timestamps
    """
    __abstract__ = True
    __table_args__ = {"extend_existing": True}

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
    created_at = sa.Column(UTCDateTime, default=now)
    updated_at = sa.Column(UTCDateTime, onupdate=now)




from arrow import Arrow
import datetime


def friendly_time(d: datetime.datetime) -> str:
    """Return "minutes ago" style date"""
    ad = Arrow.fromdatetime(d)
    other = Arrow.fromdatetime(datetime.datetime.utcnow())
    return ad.humanize(other)



from logging import Logger
from tqdm import tqdm

from sto.ethereum.txservice import EthereumStoredTXService
from sto.ethereum.utils import check_good_private_key, create_web3
from sto.models.implementation import BroadcastAccount, PreparedTransaction
from eth_account import Account
from eth_utils import to_bytes, from_wei
from sqlalchemy.orm import Session
from web3 import Web3, HTTPProvider


def get_last_transactions(logger: Logger,
              dbsession: Session,
              network: str,
              limit: int,
              ethereum_node_url: str,
              ethereum_private_key: str,
              ethereum_gas_limit: str,
              ethereum_gas_price: str,
):
    """Issue out a new Ethereum token."""

    check_good_private_key(ethereum_private_key)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    last_txs = service.get_last_transactions(limit)
    if last_txs.count() == 0:
        logger.info("No transactions yet")
        return []

    return list(last_txs)



from logging import Logger
from tqdm import tqdm

from sto.ethereum.txservice import EthereumStoredTXService
from sto.ethereum.utils import check_good_private_key, create_web3
from sto.models.implementation import BroadcastAccount, PreparedTransaction
from eth_account import Account
from eth_utils import to_bytes, from_wei
from sqlalchemy.orm import Session
from typing import Union
from web3 import Web3, HTTPProvider


def update_status(logger: Logger,
              dbsession: Session,
              network: str,
              ethereum_node_url: Union[str, Web3],
              ethereum_private_key: str,
              ethereum_gas_limit: str,
              ethereum_gas_price: str,
):
    """Issue out a new Ethereum token."""

    check_good_private_key(ethereum_private_key)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    unfinished_txs = service.get_unmined_txs()

    logger.info("Updating status for %d unfinished transactions for broadcasting in network %s", unfinished_txs.count(), network)

    if unfinished_txs.count() == 0:
        logger.info("No transactions to update. Use sto tx-last command to show the status of the last transactions.")
        return []

    unfinished_txs = list(unfinished_txs)

    # https://stackoverflow.com/questions/41985993/tqdm-show-progress-for-a-generator-i-know-the-length-of
    for tx in tqdm(unfinished_txs):
        service.update_status(tx)
        dbsession.commit()  # Try to minimise file system sync issues

    return unfinished_txs



from logging import Logger

import colorama
from sqlalchemy.orm import Session
from typing import Optional

from tqdm import tqdm

from sto.ethereum.scanner import TokenScanner
from sto.ethereum.utils import get_abi, create_web3

from sto.models.implementation import TokenScanStatus, TokenHolderDelta,TokenHolderAccount


def token_scan(logger: Logger,
              dbsession: Session,
              network: str,
              ethereum_node_url: str,
              ethereum_abi_file: Optional[str],
              token_address: str,
              start_block: Optional[int]=None,
              end_block: Optional[int]=None) -> dict:
    """Command line entry point to scan token network for events.

    By giving a block range in the middle of existing scanned range you can potentially screw up internal accounting.

    :param start_block: Block from where start scanning. If not given start from the first block or minus then of the last scanned block minus ten blocks.
    :param end_block: Block to where stop scanning. If not given scan to the latest mined block.
    :return: Mapping of address -> final amount of all addresses that were touched during the block range
    """

    abi = get_abi(ethereum_abi_file)

    web3 = create_web3(ethereum_node_url)

    scanner = TokenScanner(logger, network, dbsession, web3, abi, token_address, TokenScanStatus, TokenHolderDelta, TokenHolderAccount)

    if start_block is None:
        start_block = scanner.get_suggested_scan_start_block()

    if end_block is None:
        end_block = scanner.get_suggested_scan_end_block()

    last_scanned_block = scanner.get_last_scanned_block()

    status = scanner.get_or_create_status()
    token_name = status.name

    last_block = web3.eth.blockNumber
    logger.info("Scanning token: %s%s%s", colorama.Fore.LIGHTGREEN_EX, token_name, colorama.Fore.RESET)
    logger.info("Current last block for chain %s: %s%s%s", network, colorama.Fore.LIGHTGREEN_EX, last_block, colorama.Fore.RESET)
    logger.info("Scanning blocks: %s%d%s - %s%d%s", colorama.Fore.LIGHTGREEN_EX, start_block, colorama.Fore.RESET, colorama.Fore.LIGHTGREEN_EX, end_block, colorama.Fore.RESET)

    if last_scanned_block:
        logger.info("Last scan ended at block: %s%d%s", colorama.Fore.LIGHTGREEN_EX, last_scanned_block, colorama.Fore.RESET)

    total = end_block - start_block
    with tqdm(total=total) as progress_bar:
        def _update_progress(start, end, current, chunk_size):
            progress_bar.set_description("Scanning block: {}, batch size: {}".format(current, chunk_size))
            progress_bar.update(chunk_size)

        result = scanner.scan(start_block, end_block, progress_callback=_update_progress)
    return result


from logging import Logger

from sto.friendlytime import pretty_date
from tqdm import tqdm

from sto.ethereum.txservice import EthereumStoredTXService
from sto.ethereum.utils import check_good_private_key, create_web3
from sto.models.implementation import BroadcastAccount, PreparedTransaction
from eth_account import Account
from eth_utils import to_bytes, from_wei
from sqlalchemy.orm import Session
from web3 import Web3, HTTPProvider



class HistoryDeleteNeeded(Exception):
    pass



def restart_nonce(logger: Logger,
              dbsession: Session,
              network: str,
              ethereum_node_url: str,
              ethereum_private_key: str,
              ethereum_gas_limit: str,
              ethereum_gas_price: str,
):
    check_good_private_key(ethereum_private_key)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    service.ensure_accounts_in_sync()

    account = service.get_or_create_broadcast_account()
    txs = service.get_last_transactions(limit=1)
    if txs.count() > 0:
        raise HistoryDeleteNeeded("Cannot reset nonce as the database contains txs for {}. Delete database to restart.".format(service.address))

    # read nonce from the network and record to the database
    tx_count = web3.eth.getTransactionCount(service.address)
    account.current_nonce = tx_count

    logger.info("Address %s, nonce is now set to %d", service.address, account.current_nonce)
    dbsession.commit()


def next_nonce(logger: Logger,
              dbsession: Session,
              network: str,
              ethereum_node_url: str,
              ethereum_private_key: str,
              ethereum_gas_limit: str,
              ethereum_gas_price: str,
):
    check_good_private_key(ethereum_private_key)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)
    account = service.get_or_create_broadcast_account()
    ft = pretty_date(account.created_at)
    logger.info("Address %s, created at %s, nonce is now set to %d", service.address, ft, account.current_nonce)


"""Helper models to get token balances in the past (point of time, block num).


Please note that this implementation is reference only, SQLite compatible.
Efficiency can be greatly increased by

- Create a table Blocks which holds the reference to scanned blocks - drop blocks if they are forked

- Using binary storage for addresses, transaction hashes and such

- Using native uint256 column for values

- Calculate holder balances on the database side instead of iterating them on the application side
"""
import datetime
from binascii import hexlify
from typing import Optional, Tuple, Dict, Iterable

import sqlalchemy as sa
from decimal import Decimal

from sqlalchemy.orm import Query, object_session
from sqlalchemy.orm.attributes import flag_modified

from sto.models.utils import TimeStampedBaseModel, UTCDateTime, now


class _TokenScanStatus(TimeStampedBaseModel):
    """Hold the information of which blocks we have scanned for a certain token.

    """

    NULL_ADDRESS = "0x0000000000000000000000000000000000000000"  # Used as an address in new issuances when "from" is not available

    __tablename__ = "token_scan_status"

    #: Which network e.g. "ethereum", "kovan"
    network = sa.Column(sa.String(256), nullable=False)

    #: Address of the token contract, as hex string 0x00000, Ethereum checksummed
    address = sa.Column(sa.String(256), nullable=False)

    #: First block that was scanned (usually when the contract was deployed)
    start_block = sa.Column(sa.Integer, nullable=True)

    #: End block
    end_block = sa.Column(sa.Integer, nullable=True)

    #: When the end block was timestamped
    end_block_timestamp = sa.Column(UTCDateTime, nullable=True)

    #: Token name
    name = sa.Column(sa.String(256), nullable=True)

    #: Ticker symbol
    symbol = sa.Column(sa.String(256), nullable=True)

    #: All token balances are stored in raw amounts
    decimals = sa.Column(sa.Integer, nullable=False, default=0)

    #: Total supply, as stringified decimal
    total_supply = sa.Column(sa.String(256), nullable=True)

    def get_accounts(self, include_empty=False) -> Query:
        q = self.accounts
        if include_empty:
            pass
        else:
            q = q.filter_by(empty=False)
        return q

    def get_total_token_holder_count(self, include_empty=False):
        """How many addresses are/have been holding this token."""
        return self.get_accounts(include_empty).count()

    def get_or_create_account(self, token_holder: str) -> "_TokenHolderAccount":
        """Denormalize the token balance.

        Drop in a PostgreSQL implementation here using native databae types.
        """
        assert token_holder.startswith("0x")

        TokenHolderAccount = self.accounts.attr.target_mapper.class_

        account = self.accounts.filter_by(address=token_holder).one_or_none()
        if not account:
            account = TokenHolderAccount(address=token_holder)
            self.accounts.append(account)
        return account

    def create_deltas(self, block_num: int, block_when: datetime, txid: str, idx: int, from_: str, to_: str, value: int, TokenHolderDelta: type):
        """Creates token balance change events in the database.

        For each token transfer we create debit and credit events, so that we can nicely sum the total balance of the account.

        :param block_num: Block number
        :param block_when: Block timestamp
        :param txid: Transaction hash
        :param idx: Log index within the transaction
        :param from_: Debit account
        :param to_: Credit account
        :param value: uint256 transfer value
        """
        assert txid.startswith("0x")
        assert from_.startswith("0x")
        assert to_.startswith("0x")

        TokenHolderAccount = self.accounts.attr.target_mapper.class_

        existing = TokenHolderDelta.get_all_deltas(self).filter_by(block_num=block_num, tx_internal_order=idx).first()
        if existing:
            raise RuntimeError("Had already existing imported event: {}".format(existing))

        credit_account = self.get_or_create_account(to_)
        delta_credit = TokenHolderDelta(block_num=block_num, txid=txid, tx_internal_order=idx, block_timestamped_at=block_when)
        delta_credit.set_delta_uint(value, +1)
        credit_account.add_delta(delta_credit)

        if from_ != self.NULL_ADDRESS:
            debit_account = self.get_or_create_account(from_)
            delta_debit = TokenHolderDelta(block_num=block_num, txid=txid, tx_internal_order=idx, block_timestamped_at=block_when)
            delta_debit.set_delta_uint(value, -1)
            debit_account.add_delta(delta_debit)

    def get_raw_balance(self, address) -> int:
        """Get uint256 token balance of an address."""
        account = self.get_or_create_account(address)
        return account.get_balance_uint()

    def get_raw_balances(self, addresses: Iterable[str]) -> Dict[str, int]:
        """Get address -> balance mappings"""
        return {a: self.get_raw_balance(a) for a in addresses}

    def update_denormalised_balances(self):
        """Calculate new balance on all accounts that have been marked dirty since the last scan."""
        for account in self.accounts.filter_by(balance_calculated_at=None):
            account.update_denormalised_balance()


class _TokenHolderDelta(TimeStampedBaseModel):
    """Hold the information of which blocks we have scanned for a certain token.

    Each token transfer creates two delta events
    - Credit event: to account, positive balance change
    - Debit event: from account, negative balance change

    This is an SQLite specific hack, as it does not support 256 bit integers or true decimal types.=
    """

    __tablename__ = "token_holder_delta"

    #: First block that was scanned (usually when the contract was deployed)
    block_num = sa.Column(sa.Integer, nullable=True)

    #: When the block was timestamped
    block_timestamped_at = sa.Column(UTCDateTime, nullable=True)

    #: Give us direct link to this transaction
    txid = sa.Column(sa.String(256), nullable=True)

    #: Order of this event within the transaction, as one transaction may trigger multiple Transfer events within smart contracts
    tx_internal_order = sa.Column(sa.Integer, nullable=True)

    #: Raw uint256 data
    raw_delta = sa.Column(sa.Binary(32), nullable=False)

    #: Because raw values are uint256 and we deal with deltas, need to store the sign here. Either +1 or -1
    sign = sa.Column(sa.SmallInteger, nullable=False)

    def __str__(self):
        return "<Token:{}, address:{}, block:{}, tx:{} delta:{}>".format(self.token.address, self.address, self.block_num, self.txid, self.get_delta_uint())

    def get_delta_uint(self):
        """Return the delta as Python unlimited integer."""
        return int(hexlify(self.raw_delta), 16) * self.sign

    def set_delta_uint(self, val: int, sign: int):
        """Return the delta as Python """
        assert type(val) == int
        b = val.to_bytes(32, byteorder="big")
        self.raw_delta = b
        self.sign = sign

    @classmethod
    def get_all_deltas(cls, status: _TokenScanStatus) -> Query:
        """Get all deltas related to this token."""
        return status.accounts.join(cls)

    @classmethod
    def delete_potentially_forked_block_data(cls, status: _TokenScanStatus, after_block: int):
        """Get all deltas related to this token."""

        # TODO: Not sure how this can be offloaded 100% SQL
        session = object_session(status)
        for acc in status.accounts:
            acc.deltas.filter(cls.block_num >= after_block).delete()
            acc.mark_dirty()


class _TokenHolderAccount(TimeStampedBaseModel):
    """Hold the information of which blocks we have scanned for a certain token.

    .. note ::

        We use terms account and address interchangeable in the context of Ethereum.

    """

    __tablename__ = "token_holder_account"

    #: Address of the token contract, as hex string 0x00000
    address = sa.Column(sa.String(256), nullable=False)

    #: Denormalised balance on this account - Raw uint256 data
    #: Calculated from account deltas.
    raw_balance = sa.Column(sa.Binary(32), nullable=True)

    #: Because raw values are uint256 and we deal with deltas, need to store the sign here. Either +1 or -1.
    #: Some ERC-20 balances might go negative if we cannot correctly detect non-standard mint events.
    sign = sa.Column(sa.SmallInteger, nullable=True)

    #: SQLite hack to be able to sort balances
    sortable_balance = sa.Column(sa.Integer, nullable=True)

    #: Because sqlite cannot query uint256, we have this hack to query zero balances
    empty = sa.Column(sa.Boolean, nullable=False, default=True)

    #: When this account saw transfers last time
    last_block_num = sa.Column(sa.Integer, nullable=True)

    #: When the last transfer was timestamped
    last_block_updated_at = sa.Column(UTCDateTime, nullable=True)

    #: When the denormalised balance was calculated last time
    #: Set to null when new deltas arrive
    balance_calculated_at = sa.Column(UTCDateTime, nullable=True, default=None)

    def __str__(self):
        return "<Token:{}, holder:{}, updated at:{}, balance:{}>".format(self.token.address, self.address, self.last_updated_block, self.get_balance_uint())

    def mark_dirty(self):
        """The raw_balance does not """
        self.balance_calculated_at = None

    def is_dirty(self):
        return self.balance_calculated_at is None

    def add_delta(self, delta: _TokenHolderDelta):
        self.deltas.append(delta)
        self.mark_dirty()

    def get_balance_uint(self):
        """Return the delta as Python unlimited integer."""

        if self.is_dirty():
            raise TypeError("You need to calculate denormalised balance first")

        return int(hexlify(self.raw_balance), 16) * self.sign

    def set_balance_uint(self, val: int):
        """Return the delta as Python """

        assert type(val) == int
        b = abs(val).to_bytes(32, byteorder="big")

        self.raw_balance = b
        self.sign = -1 if val < 0 else 1

        if val != 0:
            self.empty = False
        else:
            self.empty = True

        # A hack because SQLite does not support decimals or uint256
        # TODO: Assuming 18 decimals always
        self.sortable_balance = int(Decimal(val) / Decimal(10 ** 18))

    def get_decimal_balance(self) -> Decimal:
        """Get balance in human readable decimal fractions."""
        raw_balance = self.get_balance_uint()
        return Decimal(raw_balance) / (Decimal(10) ** Decimal(self.token.decimals))

    def calculate_sum_from_deltas(self) -> Tuple[int, int, datetime.datetime]:
        """Denormalize the token balance.

        Drop in a more efficient PostgreSQL implementation here using native database types.
        """
        sum = last_block_num = 0
        last_updated_at = None

        TokenHolderDelta = self.deltas.attr.target_mapper.class_

        deltas = self.deltas.order_by(TokenHolderDelta.block_num, TokenHolderDelta.tx_internal_order)
        for d in deltas:
            sum += d.get_delta_uint()
            last_block_num = d.block_num
            last_updated_at = d.block_timestamped_at
        return sum, last_block_num, last_updated_at

    def update_denormalised_balance(self):
        """Denormalise the balance for this account."""
        raw_balance, last_block_num, last_block_at = self.calculate_sum_from_deltas()
        self.set_balance_uint(raw_balance)
        self.last_block_updated_at = last_block_at
        self.last_block_num = last_block_num
        self.balance_calculated_at = now()








import datetime
import time
from logging import Logger

from decimal import Decimal
from eth_utils import to_checksum_address
from sqlalchemy.orm import Session
from typing import Set, Dict, Tuple, Optional, Callable
from web3 import Web3
from web3.contract import Contract

from sto.ethereum.utils import getLogs
from sto.models.tokenscan import _TokenHolderAccount, _TokenScanStatus


class TokenScanner:
    """Scan blockchain for token transfer events and build a database of balances at certain timepoints (blocks)."""

    #: How far back in the past we jump to detect works in incremental rescans
    NUM_BLOCKS_RESCAN_FOR_FORKS = 10

    def __init__(self, logger: Logger, network: str, dbsession: Session, web3: Web3, abi: dict, token_address: str, TokenScanStatus: type, TokenHolderDelta: type, TokenHolderAccount: type):

        assert isinstance(web3, Web3)

        self.logger = logger
        self.network = network  # "kovan"
        self.dbsession = dbsession
        self.web3 = web3
        self.token_address = token_address
        self.abi = abi   # We need to know ERC20Standard

        # SQLAlchemy models, allow caller to supply their own
        self.TokenScanStatus = TokenScanStatus  #: type sto.models.implementation.TokenScanStatus
        self.TokenHolderDelta = TokenHolderDelta #: type sto.models.implementation.TokenHolderDelta
        self.TokenHolderAccount = TokenHolderAccount #: type sto.models.implementation.TokenHolderAccount]

        # What is the minimim
        self.min_scan_chunk_size = 10 # 12 s/block = 120 seconds period

        # This should not exceed the reply size where Infura timeouts when it tries to dump
        # Tranfer events over this many blocks
        if network == "kovan":
            # Hack to speed up test token scans
            self.max_scan_chunk_size = 500000
        else:
            self.max_scan_chunk_size = 10000

        # Factor how fast we increase the chunk size if results are found
        # # (slow down scan after starting to get hits)
        self.chunk_size_decrease = 0.5

        # Factor how was we increase chunk size if no results found
        self.chunk_size_increase = 5.0

    @property
    def address(self):
        return self.token_address

    def get_or_create_status(self) -> _TokenScanStatus:
        assert self.address.startswith("0x")
        assert self.network in ("kovan", "ethereum", "testing", "ropsten")  # TODO: Sanity check - might want to remove this

        account = self.dbsession.query(self.TokenScanStatus).filter_by(network=self.network, address=self.address).one_or_none()
        if not account:
            account = self.TokenScanStatus(network=self.network, address=self.address)
            self.dbsession.add(account)
            self.dbsession.flush()

        return account

    def update_token_info(self):
        """Update token data."""
        name, symbol, decimals, token_supply = self.get_token_contract_info(self.address)
        status = self.get_or_create_status()
        status.name = name
        status.symbol = symbol
        status.decimals = decimals
        status.total_supply = str(Decimal(token_supply) / Decimal(10 ** decimals))

    def get_contract_proxy(self, contract_name: str, address: str) -> Contract:
        """Get web3.Contract to interact directly with the network"""
        abi_data = self.abi[contract_name]

        contract_class = Contract.factory(
            web3=self.web3,
            abi=abi_data["abi"],
            bytecode=abi_data["bytecode"],
            bytecode_runtime=abi_data["bytecode_runtime"],
            )

        return contract_class(address=to_checksum_address(address))

    def get_token_contract(self, address) -> Contract:
        return self.get_contract_proxy("SecurityToken", address)

    def get_token_contract_info(self, token_address) -> [str, str, int, int]:
        """Ask token contract decimal amount using web3 and ABI."""
        contract = self.get_token_contract(token_address)
        return [contract.functions.name().call(), contract.functions.symbol().call(), contract.functions.decimals().call(), contract.functions.totalSupply().call()]

    def get_block_timestamp(self, block_num) -> datetime.datetime:
        """Get Ethereum block timestamp"""
        block_info = self.web3.eth.getBlock(block_num)
        last_time = block_info["timestamp"]
        return datetime.datetime.utcfromtimestamp(last_time)

    def get_suggested_scan_start_block(self):
        """Get where we should start to scan for new token events.

        If there are no prior scans start from block 1.
        Otherwise start from the last end block minus ten blocks.
        We rescan the last ten scanned blocks in the case there were forks to avoid
        misaccounting due to minor single block works (happens once in a hour in Ethereum).
        These heurestics could be made more robust, but this is for the sake of simple reference implementation.
        """
        status = self.get_or_create_status()
        if status.end_block:
            return max(1, status.end_block - TokenScanner.NUM_BLOCKS_RESCAN_FOR_FORKS)
        return 1

    def get_suggested_scan_end_block(self):
        """Get the last mined block."""
        return self.web3.eth.blockNumber

    def get_last_scanned_block(self):
        status = self.get_or_create_status()
        return status.end_block

    def delete_potentially_forked_block_data(self, after_block: int):
        """Purge old data in the case of a rescan."""
        status = self.get_or_create_status()
        self.TokenHolderDelta.delete_potentially_forked_block_data(status, after_block)

    def get_or_create_account(self, token_holder: str) -> _TokenHolderAccount:
        """Denormalize the token balance.

        Drop in a PostgreSQL implementation here using native databae types.
        """
        assert token_holder.startswith("0x")
        status = self.get_or_create_status()
        return status.get_or_create_account(token_holder)

    def create_deltas(self, block_num: int, block_when: datetime, txid: str, idx: int, from_: str, to_: str, value: int):
        """Creates token balance change events in the database.

        For each token transfer we create debit and credit events, so that we can nicely sum the total balance of the account.
        """
        status = self.get_or_create_status()
        status.create_deltas(block_num, block_when, txid, idx, from_, to_, value, self.TokenHolderDelta)
        self.dbsession.flush()

    def scan_chunk(self, start_block, end_block) -> Set[str]:
        """Populate TokenHolderStatus for certain blocks.

        :return: Set of addresses where balance changes between scans.
        """

        mutated_addresses = set()
        token = self.get_token_contract(self.address)

        # Discriminate between ERC-20 transfer and ERC-667
        # The latter is not used anywhere yet AFAIK
        Transfer = token.events.Transfer("from", "to", "value")
        Issued = token.events.Issued("to", "value")
        block_timestamps = {}
        get_block_timestamp = self.get_block_timestamp

        # Cache block timestamps to reduce some RPC overhead
        # Real solution would be smarter models around block
        def get_block_when(block_num):
            if not block_num in block_timestamps:
                block_timestamps[block_num] = get_block_timestamp(block_num)
            return block_timestamps[block_num]

        for event_type in [Issued, Transfer]:

            # events = event_type.createFilter(fromBlock=start_block, toBlock=end_block).get_all_entries()

            events = getLogs(event_type, fromBlock=start_block, toBlock=end_block)

            # AttributeDict({'args': AttributeDict({'from': '0xDE5bC059aA433D72F25846bdFfe96434b406FA85', 'to': '0x0bdcc26C4B8077374ba9DB82164B77d6885b92a6', 'value': 300000000000000000000}), 'event': 'Transfer', 'logIndex': 0, 'transactionIndex': 0, 'transactionHash': HexBytes('0x973eb270e311c23dd6173a9092c9ad4ee8f3fe24627b43c7ad75dc2dadfcbdf9'), 'address': '0x890042E3d93aC10A426c7ac9e96ED6416B0cC616', 'blockHash': HexBytes('0x779f55173414a7c0df0d9fc0ab3fec461a66ceeee0b4058e495d98830c92abf8'), 'blockNumber': 7})
            for e in events:
                idx = e["logIndex"]  # nteger of the log index position in the block. null when its pending log.
                if idx is None:
                    raise RuntimeError("Somehow tried to scan a pending block")

                if e["event"] == "Issued":
                    # New issuances pop up from empty air - mark this specially in the database.
                    # Also some ERC-20 tokens use Transfer from null address to symbolise issuance.
                    from_ = self.TokenScanStatus.NULL_ADDRESS
                else:
                    from_ = e["args"]["from"]
                    mutated_addresses.add(e["args"]["from"])

                block_when = get_block_when(e["blockNumber"])

                self.create_deltas(e["blockNumber"], block_when, e["transactionHash"].hex(), idx, from_, e["args"]["to"], e["args"]["value"])
                self.logger.debug("Imported %s, token:%s block:%d from:%s to:%s value:%s", e["event"], self.address, e["blockNumber"], from_, e["args"]["to"], e["args"]["value"])

                mutated_addresses.add(e["args"]["to"])

        return mutated_addresses

    def estimate_next_chunk_size(self, current_chuck_size: int, event_found_count: int):
        """Try to figure out optimal chunk size

        Our scanner might need to scan the whole blockchain for all events

        * We want to minimize API calls over empty blocks

        * We want to make sure that one scan chunk does not try to process too many entries once, as we try to control commit buffer size and potentially asynchronous busy loop

        * Do not overload node serving JSON-RPC API

        TODO: eth_getLogs does not provide meaningful way to get the block number when events start. Make a feature request. Somewhere.

        This heurestics exponentially increases and the scan chunk size depending on if we are seeing events or not.
        When any transfers are encountered we are back to scan only few blocks at the time.
        It does not make sense to do a full chain scan starting from block 1, doing one JSON-RPC call per 20 blocks.
        """

        if event_found_count > 0:
            # When we encounter first events then reset the chunk size window
            current_chuck_size = self.min_scan_chunk_size
        else:
            current_chuck_size *= self.chunk_size_increase

        current_chuck_size = max(self.min_scan_chunk_size, current_chuck_size)
        current_chuck_size = min(self.max_scan_chunk_size, current_chuck_size)
        return int(current_chuck_size)

    def update_scan_status(self, start_block, end_block):
        # Update token scan status
        status = self.get_or_create_status()
        if not status.start_block:
            status.start_block = start_block

        status.end_block = end_block
        status.end_block_timestamp = self.get_block_timestamp(status.end_block)

    def scan(self, start_block, end_block, start_chunk_size=20, progress_callback=Optional[Callable]) -> dict:
        """Perform a token balances scan.

        Assumes all balances in the database are valid before start_block (no forks sneaked in).

        :param start_block: The first block included in the scan

        :param end_block: The last block included in the scan

        :return: Address -> last balance mapping for all address balances that changed during those blocks
        """

        assert start_block <= end_block

        self.delete_potentially_forked_block_data(start_block)
        self.update_token_info()

        current_block = start_block
        updated_token_holders = set()  # Token holders that get updates

        # Scan in chunks, commit between
        chunk_size = start_chunk_size
        mutated_addresses = set()
        last_scan_duration = last_logs_found = 0
        status = self.get_or_create_status()

        while current_block <= end_block:

            # Where does our current chunk scan ends - are we out of chain yet?
            current_end = min(current_block + chunk_size, end_block)

            # Print some diagnostics to logs to try to fiddle with real world JSON-RPC API performance
            self.logger.debug("Scanning token transfers for blocks: %d - %d, chunk size %d, last chunk scan took %f, last logs found %d", current_block, current_end, chunk_size, last_scan_duration, last_logs_found)
            start = time.time()

            # Process blocks for this token over eth_getLogs
            mutated_addresses = self.scan_chunk(current_block, current_end)
            last_scan_duration = time.time() - start
            last_logs_found = len(mutated_addresses)

            # Manage the list of what addresses our scan has touched
            updated_token_holders.update(mutated_addresses)

            # Persistent the state how are along we are in the scan
            self.update_scan_status(start_block, current_end)

            # Update database on the disk
            self.dbsession.commit()

            # Print progress bar
            if progress_callback:
                progress_callback(start_block, end_block, current_block, chunk_size)

            # Try to guess who many blocks we try to fetch over eth_getLogs API next time
            chunk_size = self.estimate_next_chunk_size(chunk_size, len(mutated_addresses))

            # Set where the next chunk starts
            current_block = current_end + 1

        # Calculate balances to all accounts that have not seen new total since the last scan
        status.update_denormalised_balances()
        self.dbsession.commit()  # Write latest balances

        result = status.get_raw_balances(mutated_addresses)
        return result




import re
import textwrap
from click import Group, Context


def remove_ansi(s):
    """https://stackoverflow.com/a/38662876/315168"""
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', s)


def generate_reference(cli: Group):
    with cli.make_context("sto", ["sto"], max_content_width=200, terminal_width=200) as ctx:
        main_help = cli.get_help(ctx)
        main_help = remove_ansi(main_help)
        main_help = textwrap.indent(main_help, "   ")
        print(TEMPLATE.format(main_help))

        for name in sorted(cli.commands.keys()):
            cmd = cli.commands[name]
            with cli.make_context("sto {}".format(name), ["sto", name]) as subcommand_ctx:
                short_help = cmd.help
                long_help = cmd.get_help(subcommand_ctx)
                long_help = textwrap.indent(long_help, "    ")
                print(SUBCOMMAND_TEMPLATE.format(name, name, short_help, long_help))

TEMPLATE = """
Command line reference
======================

Here is the command line reference for ``sto`` command.

.. contents:: :local:

Options and config files
------------------------

Settings can be either given in a config file, specified by ``--config-file`` switch or directly to the main command.

E.g. these are equivalent.

Command line:

.. code-block:: shell

    sto --ethereum-node-url="https://mainnet.infura.io/v3/453d2049c15d4a8da5501a0464fa44f8" token-scan ...
    
As with INI file ``mainnet.ini``:

.. code-block:: ini
    
    # Infura mainnet net node url
    ethereum-node-url = https://mainnet.infura.io/v3/453d2049c15d4a8da5501a0464fa44f8
    
.. code-block:: shell

    sto --config-file=mainnet.ini token-scan ...

Subcommands take their own options that cannot be specified in the settings file.
 
Main command and options
------------------------

When running ``sto --help`` you get list of settings and subcommands:

.. code-block:: text

{}
"""


SUBCOMMAND_TEMPLATE = """

.. _{}:

{}
-------------------------------------

{}

.. code-block:: text

{}

"""


import binascii
import calendar
import datetime
from logging import Logger
from typing import Optional

from eth_account import Account
from eth_utils import from_wei, to_bytes
from web3 import Web3, HTTPProvider
from sto.ethereum.utils import check_good_node_url, check_good_private_key, create_web3


class NodeNotSynced(Exception):
    pass


class NeedMoney(Exception):
    pass


def diagnose(logger: Logger, node_url: str, private_key_hex: str) -> Optional[Exception]:
    """Run Ethereum connection and account diagnostics.

    Check that the user has properly configured Ethereum node and private key.

    Never fails. Exceptions are written to the logger output and returned.
    """

    try:
        check_good_node_url(node_url)

        logger.info("Attempting to connect to Ethereum node %s", node_url)
        web3 = create_web3(node_url)

        logger.info("Connected to Ethereum node software %s", web3.version.node)

        block_num = web3.eth.blockNumber

        d = datetime.datetime.utcnow()
        unix_time = calendar.timegm(d.utctimetuple())

        block_info = web3.eth.getBlock(block_num)
        last_time = block_info["timestamp"]

        if last_time == 0:
            raise NodeNotSynced("Looks like your node has not yet been synced.")

        ago = unix_time - last_time

        logger.info("Current Ethereum node block number: %d, last block %d seconds ago - compare this to data on https://etherscan.io", block_num, ago)

        if ago < 0:
            raise NodeNotSynced("Last block in the future? Do we have a clock with a wrong timezone somewhere?")

        if ago > 1800:
            raise NodeNotSynced("Looks like your node has not received a block for half an hour. It is most likely unsynced at the moment.")

        check_good_private_key(private_key_hex)

        logger.info("Using private key %s...", private_key_hex[0:3])
        account = Account.privateKeyToAccount(to_bytes(hexstr=private_key_hex))

        balance = web3.eth.getBalance(account.address)
        logger.info("Address %s has ETH balance of %f", account.address, from_wei(balance, "ether"))

        if balance == 0:
            raise NeedMoney("Your Ethereum account {} needs to have ETH in order to use this tool".format(account.address))


    except Exception as e:

        logger.error("Diagnostics failure")
        logger.exception(e)
        return e


"""Define command line interface and subcommands. """
import logging
import os
import sys

import colorama
import configobj
import pkg_resources

import click
import coloredlogs
from sto.db import setup_database


class UnknownConfiguredNetwork(Exception):
    pass


class BoardCommmadConfiguration:
    """All top level option switcs either read command line or INI

    See ``main.cli()`` arguments for content.
    """

    def __init__(self, **kwargs):
        # See cli() for descriptions of variables
        self.__dict__.update(kwargs)


def create_command_line_logger(log_level):
    """Create a fancy output.

    See: https://coloredlogs.readthedocs.io/en/latest/readme.html#installation
    """
    fmt = "%(message)s"
    logger = logging.getLogger()
    coloredlogs.install(level=log_level, fmt=fmt, logger=logger)
    return logger


def is_ethereum_network(network: str):
    return network in ("ethereum", "kovan", "ropsten")



INTRO_TEXT = """{}TokenMarket{} security token management tool.

    {}Manage tokenised equity for things like issuing out new, distributing and revoking shares.{}
    
    For full documentation see https://docs.tokenmarket.net/
""".format(colorama.Fore.LIGHTGREEN_EX, colorama.Fore.RESET, colorama.Fore.BLUE, colorama.Fore.RESET)



@click.group(help=INTRO_TEXT)
@click.option('--config-file', required=False, default=None, help="INI file where to read options from", type=click.Path())
@click.option('--database-file', required=False, default="transactions.sqlite", help="SQLite file that persists transaction broadcast status", type=click.Path())
@click.option('--network', required=False, default="ethereum", help="Network name. Either 'ethereum' or 'kovan' are supported for now.")
@click.option('--ethereum-node-url', required=False, default="http://localhost:8545", help="Parity or Geth JSON-RPC to connect for Ethereum network access")
@click.option('--ethereum-abi-file', required=False, help='Solidity compiler output JSON to override default smart contracts')
@click.option('--ethereum-gas-price', required=False, help='How many GWei we pay for gas')
@click.option('--ethereum-gas-limit', required=False, help='What is the transaction gas limit for broadcasts', type=int)
@click.option('--ethereum-private-key', required=False, help='Private key for the broadcasting account')
@click.option('--etherscan-api-key', required=False, help='EtherScan API key used for the contract source code verification')
@click.option('--log-level', default="INFO", help="Python logging level to tune the verbosity of the command")
@click.option('--auto-restart-nonce', default=True, help="Automatically restart nonce for the deployment account if starting with a fresh database", type=bool)
@click.pass_context
def cli(ctx, config_file, **kwargs):

    # Fill in arguments from the configuration file
    if config_file:
        if not os.path.exists(config_file):

            sys.exit("Config file does not exist {}".format(config_file))

        config = configobj.ConfigObj(config_file, raise_errors=True)

        # TODO: Bug here - could not figure out how to pull out from click if an option is set on a command line or are we using default.
        # Thus you cannot override config file variables by giving a default value from command line
        for opt in ctx.command.params:  # type: click.core.Options:

            dashed_name = opt.name.replace("_", "-")
            value = kwargs.get(opt.name)
            if value == opt.default:
                config_file_value = config.get(dashed_name)
                if config_file_value:
                    kwargs[opt.name] = config_file_value  # TODO: opt.process_value

    log_level = kwargs["log_level"]

    config = BoardCommmadConfiguration(**kwargs)
    logger = config.logger = create_command_line_logger(log_level.upper())

    # Mute SQLAlchemy logger who is quite a verbose friend otherwise
    sa_logger = logging.getLogger("sqlalchemy")
    sa_logger.setLevel(logging.WARN)

    # TODO: No idea how to peek into click internals to figure out upcoming subcommand
    prelude = True
    if sys.argv[-1] == "version":
        # Mute generic prelude
        prelude = False

    # Print out the info
    dbfile = os.path.abspath(config.database_file)
    version = pkg_resources.require("sto")[0].version
    config.version = version

    if prelude:
        copyright = "Copyright TokenMarket Ltd. 2018 - 2019"
        logger.info("STO tool, version %s%s%s - %s", colorama.Fore.LIGHTCYAN_EX, version, colorama.Fore.RESET, copyright)
        logger.info("Using database %s%s%s", colorama.Fore.LIGHTCYAN_EX, dbfile, colorama.Fore.RESET)

        config.dbsession, new_db = setup_database(logger, dbfile)
        if new_db:
            if config.auto_restart_nonce and config.ethereum_private_key:
                logger.info("Automatically fetching the initial nonce for the deployment account from blockchain")
                from sto.ethereum.nonce import restart_nonce
                restart_nonce(logger,
                              config.dbsession,
                              config.network,
                              ethereum_node_url=config.ethereum_node_url,
                              ethereum_private_key=config.ethereum_private_key,
                              ethereum_gas_limit=config.ethereum_gas_limit,
                              ethereum_gas_price=config.ethereum_gas_price)

    ctx.obj = config


@cli.command()
@click.option('--symbol', required=True)
@click.option('--name', required=True)
@click.option('--amount', required=True, type=int)
@click.option('--transfer-restriction', required=False, default="unrestricted")
@click.pass_obj
def issue(config: BoardCommmadConfiguration, symbol, name, amount, transfer_restriction):
    """Issue out a new security token.

    * Creates a new share series

    * Allocates all new shares to the management account

    * Sets the share transfer restriction mode
    """

    logger = config.logger

    assert is_ethereum_network(config.network) # Nothing else implemented yet

    from sto.ethereum.issuance import deploy_token_contracts
    from sto.ethereum.txservice import EthereumStoredTXService

    dbsession = config.dbsession

    txs = deploy_token_contracts(logger,
                          dbsession,
                          config.network,
                          ethereum_node_url=config.ethereum_node_url,
                          ethereum_abi_file=config.ethereum_abi_file,
                          ethereum_private_key=config.ethereum_private_key,
                          ethereum_gas_limit=config.ethereum_gas_limit,
                          ethereum_gas_price=config.ethereum_gas_price,
                          name=name,
                          symbol=symbol,
                          amount=amount,
                          transfer_restriction=transfer_restriction)

    EthereumStoredTXService.print_transactions(txs)

    # Write database
    dbsession.commit()

    logger.info("Run %ssto tx-broadcast%s to write this to blockchain", colorama.Fore.LIGHTCYAN_EX, colorama.Fore.RESET)


@cli.command(name="issue-logs")
@click.pass_obj
def past_issuances(config: BoardCommmadConfiguration):
    """Print out transactions of for tokens issued in the past."""

    logger = config.logger

    from sto.ethereum.issuance import past_issuances

    dbsession = config.dbsession

    txs = list(past_issuances(config, dbsession))

    if txs:
        from sto.ethereum.txservice import EthereumStoredTXService
        EthereumStoredTXService.print_transactions(txs)
        logger.info("See column %sto%s for token contract addresses", colorama.Fore.LIGHTCYAN_EX, colorama.Fore.RESET)
    else:
        logger.info("No issuances")



@cli.command(name="token-status")
@click.option('--address', required=True, help="Token contract addrss")
@click.pass_obj
def status(config: BoardCommmadConfiguration, address):
    """Print token contract status."""

    logger = config.logger

    assert is_ethereum_network(config.network) # Nothing else implemented yet

    from sto.ethereum.issuance import contract_status

    dbsession = config.dbsession

    contract_status(logger,
      dbsession,
      config.network,
      ethereum_node_url=config.ethereum_node_url,
      ethereum_abi_file=config.ethereum_abi_file,
      ethereum_private_key=config.ethereum_private_key,
      ethereum_gas_limit=config.ethereum_gas_limit,
      ethereum_gas_price=config.ethereum_gas_price,
      token_contract=address)


@cli.command(name="distribute-multiple")
@click.option('--csv-input', required=True, help="CSV file for entities receiving tokens")
@click.option('--address', required=True, help="Token contract address")
@click.pass_obj
def distribute_multiple(config: BoardCommmadConfiguration, csv_input, address):
    """Distribute shares to multiple shareholders whose address info is read from a file."""

    logger = config.logger

    assert is_ethereum_network(config.network) # Nothing else implemented yet
    dbsession = config.dbsession

    from sto.distribution import read_csv
    from sto.ethereum.distribution import distribute_tokens

    dists = read_csv(logger, csv_input)
    if not dists:
        sys.exit("Empty CSV file")

    new_txs, old_txs = distribute_tokens(
        logger,
        dbsession,
        config.network,
        ethereum_node_url=config.ethereum_node_url,
        ethereum_abi_file=config.ethereum_abi_file,
        ethereum_private_key=config.ethereum_private_key,
        ethereum_gas_limit=config.ethereum_gas_limit,
        ethereum_gas_price=config.ethereum_gas_price,
        token_address=address,
        dists=dists,
    )

    logger.info("Distribution created %d new transactions and there was already %d old transactions in the database", new_txs, old_txs)

    # Write database
    dbsession.commit()

    logger.info("Run %ssto tx-broadcast%s to send out distribured shares to the world", colorama.Fore.LIGHTCYAN_EX, colorama.Fore.RESET)


@cli.command(name="distribute-single")
@click.option('--token-address', required=True, help="Token contract address")
@click.option('--to-address', required=True, help="Receiver")
@click.option('--external-id', required=True, help="External id string for this transaction - no duplicates allowed")
@click.option('--email', required=True, help="Receiver email (for audit log only)")
@click.option('--name', required=True, help="Receiver name (for audit log only)")
@click.option('--amount', required=True, help="Amount of tokens as a decimal number")
@click.pass_obj
def distribute_single(config: BoardCommmadConfiguration, token_address, to_address, external_id, email, name, amount):
    """Send out tokens to one individual shareholder."""

    logger = config.logger

    assert is_ethereum_network(config.network) # Nothing else implemented yet
    dbsession = config.dbsession

    from sto.ethereum.distribution import distribute_single

    result = distribute_single(
        logger,
        dbsession,
        config.network,
        ethereum_node_url=config.ethereum_node_url,
        ethereum_abi_file=config.ethereum_abi_file,
        ethereum_private_key=config.ethereum_private_key,
        ethereum_gas_limit=config.ethereum_gas_limit,
        ethereum_gas_price=config.ethereum_gas_price,
        token_address=token_address,
        to_address=to_address,
        ext_id=external_id,
        email=email,
        name=name,
        amount=amount
    )

    if result:
        # Write database
        dbsession.commit()
        logger.info("Run %ssto tx-broadcast%s to send out distribured shares to the world", colorama.Fore.LIGHTCYAN_EX, colorama.Fore.RESET)


@cli.command()
@click.pass_obj
def diagnose(config: BoardCommmadConfiguration):
    """Check your node and account status.

    This command will print out if you are properly connected to Ethereum network and your management account has enough Ether balance.
    """

    # Run Ethereum diagnostics
    if is_ethereum_network(config.network):
        from sto.ethereum.diagnostics import diagnose

        private_key = config.ethereum_private_key
        exception = diagnose(config.logger, config.ethereum_node_url, private_key)
        if exception:
            config.logger.error("{}We identified an issue with your configuration. Please fix the issue above to use this command yet.{}".format(colorama.Fore.RED, colorama.Fore.RESET))
        else:
            config.logger.info("{}Ready for action.{}".format(colorama.Fore.LIGHTGREEN_EX, colorama.Fore.RESET))
    else:
        raise UnknownConfiguredNetwork()


@cli.command(name="ethereum-create-account")
@click.pass_obj
def create_ethereum_account(config: BoardCommmadConfiguration):
    """Creates a new Ethereum account."""

    config.logger.info("Creating new Ethereum account.")
    from sto.ethereum.account import create_account_console
    create_account_console(config.logger, config.network)


@cli.command(name="tx-broadcast")
@click.pass_obj
def broadcast(config: BoardCommmadConfiguration):
    """Broadcast waiting transactions.

    Send all management account transactions to Ethereum network.
    After a while, transactions are picked up by miners and included in the blockchain.
    """

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.broadcast import broadcast

    dbsession = config.dbsession

    txs = broadcast(logger,
                          dbsession,
                          config.network,
                          ethereum_node_url=config.ethereum_node_url,
                          ethereum_private_key=config.ethereum_private_key,
                          ethereum_gas_limit=config.ethereum_gas_limit,
                          ethereum_gas_price=config.ethereum_gas_price)

    if txs:
        from sto.ethereum.txservice import EthereumStoredTXService
        EthereumStoredTXService.print_transactions(txs)
        logger.info("Run %ssto tx-update%s to monitor your transaction propagation status", colorama.Fore.LIGHTCYAN_EX, colorama.Fore.RESET)

    # Write database
    dbsession.commit()


@cli.command(name="tx-update")
@click.pass_obj
def update(config: BoardCommmadConfiguration):
    """Update transaction status.

    Connects to Ethereum network, queries the status of our broadcasted transactions.
    Then print outs the still currently pending transactions or freshly mined transactions.
    """

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.status import update_status

    dbsession = config.dbsession

    txs = update_status(logger,
                          dbsession,
                          config.network,
                          ethereum_node_url=config.ethereum_node_url,
                          ethereum_private_key=config.ethereum_private_key,
                          ethereum_gas_limit=config.ethereum_gas_limit,
                          ethereum_gas_price=config.ethereum_gas_price)

    if txs:
        from sto.ethereum.txservice import EthereumStoredTXService
        EthereumStoredTXService.print_transactions(txs)

    # Write database
    dbsession.commit()



@cli.command(name="tx-verify")
@click.pass_obj
def verify(config: BoardCommmadConfiguration):
    """Verify source code of contract deployment transactions on EtherScan.

    Users EtherScan API to verify all deployed contracts from the management account.
    """

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.issuance import verify_source_code

    dbsession = config.dbsession

    txs = verify_source_code(logger,
                          dbsession,
                          config.network,
                          config.etherscan_api_key)

    if txs:
        from sto.ethereum.txservice import EthereumStoredTXService
        EthereumStoredTXService.print_transactions(txs)

    # Write database
    dbsession.commit()


@cli.command(name="tx-last")
@click.option('--limit', required=False, help="How many transactions to print", default=5)
@click.pass_obj
def last(config: BoardCommmadConfiguration, limit):
    """Print latest transactions from database.
    """

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.last import get_last_transactions

    dbsession = config.dbsession

    txs = get_last_transactions(logger,
                          dbsession,
                          config.network,
                          limit=limit,
                          ethereum_node_url=config.ethereum_node_url,
                          ethereum_private_key=config.ethereum_private_key,
                          ethereum_gas_limit=config.ethereum_gas_limit,
                          ethereum_gas_price=config.ethereum_gas_price)

    if txs:
        from sto.ethereum.txservice import EthereumStoredTXService
        EthereumStoredTXService.print_transactions(txs)


@cli.command(name="tx-restart-nonce")
@click.pass_obj
def restart_nonce(config: BoardCommmadConfiguration):
    """Resets the broadcasting account nonce."""

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.nonce import restart_nonce

    dbsession = config.dbsession

    restart_nonce(logger,
          dbsession,
          config.network,
          ethereum_node_url=config.ethereum_node_url,
          ethereum_private_key=config.ethereum_private_key,
          ethereum_gas_limit=config.ethereum_gas_limit,
          ethereum_gas_price=config.ethereum_gas_price)

@cli.command(name="tx-next-nonce")
@click.pass_obj
def next_nonce(config: BoardCommmadConfiguration):
    """Print next nonce to be consumed."""

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.nonce import next_nonce

    dbsession = config.dbsession

    txs = next_nonce(logger,
                          dbsession,
                          config.network,
                          ethereum_node_url=config.ethereum_node_url,
                          ethereum_private_key=config.ethereum_private_key,
                          ethereum_gas_limit=config.ethereum_gas_limit,
                          ethereum_gas_price=config.ethereum_gas_price)


@cli.command(name="token-scan")
@click.option('--start-block', required=False, help="The first block where we start (re)scan", default=None)
@click.option('--end-block', required=False, help="Until which block we scan, also can be 'latest'", default=None)
@click.option('--token-address', required=True, help="Token contract address", default=None)
@click.pass_obj
def token_scan(config: BoardCommmadConfiguration, token_address, start_block, end_block):
    """Update token holder balances from a blockchain to a local database.

    Reads the Ethereum blockchain for a certain token and builds a local database of token holders and transfers.

    If start block and end block information are omitted, continue the scan where we were left last time.
    Scan operations may take a while.
    """

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.ethereum.tokenscan import token_scan

    dbsession = config.dbsession

    updated_addresses = token_scan(logger,
                          dbsession,
                          config.network,
                          ethereum_node_url=config.ethereum_node_url,
                          ethereum_abi_file=config.ethereum_abi_file,
                          token_address=token_address,
                          start_block=start_block,
                          end_block=end_block)

    logger.info("Updated %d token holder balances", len(updated_addresses))


@cli.command(name="cap-table")
@click.option('--identity-file', required=False, help="CSV file containing address real world identities", default=None, type=click.Path())
@click.option('--token-address', required=True, help="Token contract address", default=None)
@click.option('--order-by', required=False, help="How cap table is sorted", default="balance", type=click.Choice(["balance", "name", "updated", "address"]))
@click.option('--order-direction', required=False, help="Sort direction", default="desc", type=click.Choice(["asc", "desc"]))
@click.option('--include-empty', required=False, help="Sort direction", default=False, type=bool)
@click.option('--max-entries', required=False, help="Print only first N entries", default=5000, type=int)
@click.option('--accuracy', required=False, help="How many decimals include in balance output", default=2, type=int)
@click.pass_obj
def cap_table(config: BoardCommmadConfiguration, token_address, identity_file, order_by, order_direction, include_empty, max_entries, accuracy):
    """Print out token holder cap table.

    The token holder data must have been scanned earlier using token-scan command.

    You can supply optional CSV file that contains Ethereum address mappings to individual token holder names.
    """

    assert is_ethereum_network(config.network)

    logger = config.logger

    from sto.generic.captable import generate_cap_table, print_cap_table
    from sto.identityprovider import read_csv, NullIdentityProvider, CSVIdentityProvider
    from sto.models.implementation import TokenScanStatus, TokenHolderAccount

    dbsession = config.dbsession

    if identity_file:
        entries = read_csv(logger, identity_file)
        provider = CSVIdentityProvider(entries)
    else:
        provider = NullIdentityProvider()

    cap_table = generate_cap_table(logger,
                          dbsession,
                          token_address=token_address,
                          identity_provider=provider,
                          order_by=order_by,
                          order_direction=order_direction,
                          include_empty=include_empty,
                          TokenScanStatus=TokenScanStatus,
                          TokenHolderAccount=TokenHolderAccount)

    print_cap_table(cap_table, max_entries, accuracy)


@cli.command(name="reference")
@click.pass_obj
def reference(config: BoardCommmadConfiguration):
    """Print out the command line reference for the documentation."""

    from sto.generic.reference import generate_reference
    generate_reference(cli)


@cli.command(name="version")
@click.pass_obj
def version(config: BoardCommmadConfiguration):
    """Print version number and exit."""
    print(config.version)



def main():
    # https://github.com/pallets/click/issues/204#issuecomment-270012917
    cli.main(max_content_width=200, terminal_width=200)



from logging import Logger
from tqdm import tqdm

from sto.ethereum.txservice import EthereumStoredTXService
from sto.ethereum.utils import check_good_private_key, create_web3
from sto.models.implementation import BroadcastAccount, PreparedTransaction
from eth_account import Account
from eth_utils import to_bytes, from_wei
from sqlalchemy.orm import Session
from typing import Union, Optional
from web3 import Web3, HTTPProvider


def broadcast(logger: Logger,
              dbsession: Session,
              network: str,
              ethereum_node_url: Union[str, Web3],
              ethereum_private_key: str,
              ethereum_gas_limit: Optional[str],
              ethereum_gas_price: Optional[str],
):
    """Issue out a new Ethereum token."""

    check_good_private_key(ethereum_private_key)

    web3 = create_web3(ethereum_node_url)

    service = EthereumStoredTXService(network, dbsession, web3, ethereum_private_key, ethereum_gas_price, ethereum_gas_limit, BroadcastAccount, PreparedTransaction)

    service.ensure_accounts_in_sync()

    pending_broadcasts = service.get_pending_broadcasts()

    logger.info("Pending %d transactions for broadcasting in network %s", pending_broadcasts.count(), network)

    if pending_broadcasts.count() == 0:
        logger.info("No new transactions to broadcast. Use sto tx-update command to see tx status.")
        return []

    account = Account.privateKeyToAccount(ethereum_private_key)
    balance = web3.eth.getBalance(account.address)

    logger.info("Our address %s has ETH balance of %f for operations", account.address, from_wei(balance, "ether"))

    txs = list(pending_broadcasts)
    # https://stackoverflow.com/questions/41985993/tqdm-show-progress-for-a-generator-i-know-the-length-of
    for tx in tqdm(txs, total=pending_broadcasts.count()):
        try:
            service.broadcast(tx)
            # logger.info("Broadcasted %s", tx.txid)
        except Exception as e:
            logger.exception(e)
            logger.error("Failed to broadcast transaction %s: %s", tx.txid, tx.human_readable_description)
            raise e

        dbsession.commit()  # Try to minimise file system sync issues

    return txs



"""Match Ethereum addresses to real world identities."""
import csv
from abc import ABC
from decimal import Decimal
from logging import Logger

from typing import List, Optional

from eth_utils import to_checksum_address, is_checksum_address

from sto.ethereum.utils import validate_ethereum_address


class IdentityEntry:
    """Reveal real world identity behind an address.

    Used when priting out cap table.
    """

    def __init__(self, name: str, address: str):
        """
        :param name: Entity name or person name
        :param address: Crypto address where STOs are delivered
        """
        self.name = name
        self.address = address


def read_csv(logger: Logger, fname) -> List[IdentityEntry]:
    """Read identity information from a CSV file.

    Columns are:
    * name
    * address
    """

    logger.info("Reading CSV input %s", fname)
    with open(fname, "rt") as inp:
        reader = csv.DictReader(inp)
        rows = [row for row in reader]

    result = []

    # TODO: Address format is now hardcoded for Etheereum.
    for idx, row in enumerate(rows):
        addr = row["address"].strip()
        try:
            validate_ethereum_address(addr)
        except ValueError as e:
            raise ValueError("Invalid Ethereum address on row:" + str(idx+1) + " address:", addr +  " reason:" + str(e) + " name:" + row["name"]) from e

        addr = to_checksum_address(addr)
        entry = IdentityEntry(row["name"], addr)
        result.append(entry)

    return result


class IdentityProvider(ABC):
    """Allow flexible identity backend.

    We can support files, KYC databases, third party services.
    """

    def get_identity(self, address: str) -> Optional[IdentityEntry]:
        """Map address to an identity."""


class NullIdentityProvider(IdentityProvider):
    """Nobody knows who holds the shares. The cryptoanarchist dream.."""

    def __init__(self):
        pass

    def get_identity(self, address) -> Optional[IdentityEntry]:
        return None


class CSVIdentityProvider(IdentityProvider):
    """Hold identities in memory after reading from CSV."""

    def __init__(self, entries: List[IdentityEntry]):
        self.map = {}
        for entry in entries:
            self.map[entry.address] = entry

    def get_identity(self, address) -> Optional[IdentityEntry]:
        assert is_checksum_address(address)
        return self.map.get(address)






"""Attach SQLAlchemy models to their "Base"

The framework user may supply their own base class e.g. for the case of fine tuning some database options.
Here we supply implementations for the default command line application using SQLite.
"""

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base

from .broadcastaccount import _BroadcastAccount, _PreparedTransaction
from .tokenscan import _TokenScanStatus, _TokenHolderDelta, _TokenHolderAccount


Base = declarative_base()


class BroadcastAccount(_BroadcastAccount, Base):
    pass


class PreparedTransaction(_PreparedTransaction, Base):

    broadcast_account_id = sa.Column(sa.ForeignKey("broadcast_account.id"), nullable=True)
    broadcast_account = orm.relationship(BroadcastAccount,
                        backref=orm.backref("txs",
                                        lazy="dynamic",
                                        cascade="all, delete-orphan",
                                        single_parent=True, ), )


class TokenScanStatus(_TokenScanStatus, Base):
    pass


class TokenHolderAccount(_TokenHolderAccount, Base):

    token_id = sa.Column(sa.ForeignKey("token_scan_status.id"), nullable=False)
    token = orm.relationship(TokenScanStatus,
                        backref=orm.backref("accounts",
                                        lazy="dynamic",
                                        cascade="all, delete-orphan",
                                        single_parent=True, ), )


class TokenHolderDelta(_TokenHolderDelta, Base):

    account_id = sa.Column(sa.ForeignKey("token_holder_account.id"), nullable=False)
    account = orm.relationship(TokenHolderAccount,
                        backref=orm.backref("deltas",
                                        lazy="dynamic",
                                        cascade="all, delete-orphan",
                                        single_parent=True, ), )





from typing import Optional

import sqlalchemy as sa
from sqlalchemy.orm.attributes import flag_modified

from sto.models.utils import TimeStampedBaseModel, UTCDateTime


class _BroadcastAccount(TimeStampedBaseModel):
    """Account status we maintain for broadcast transactions."""

    __tablename__ = "broadcast_account"

    #: Network name like "kovan", "ethereum"
    network = sa.Column(sa.String(256), nullable=False)

    #: Address of the account as hex string, like 0x000000
    address = sa.Column(sa.String(256), nullable=False)

    #: Currently available nonce to be allocated for the next transaction
    current_nonce = sa.Column(sa.Integer, default=0)


class _PreparedTransaction(TimeStampedBaseModel):
    """Manage transactions.

    Make is safe to rebroadcast.
    """

    __tablename__ = "prepared_transaction"

    #: Ethereum transaction nonce allocated from BroadcastAccount for this transaction
    nonce = sa.Column(sa.Integer, default=1)

    #: What is the corresponding transaction for this crypto transactions in other systems. Could be payment TXID or fiat receipt id in the case of purchasing shares. Under normal circumstances PreparedTransaction should not have duplicate external_ids but this may change under rebroadcast and other manual fix ups.
    external_id = sa.Column(sa.String(256), nullable=True)

    # Is this a contract deployment transaction
    contract_deployment = sa.Column(sa.Boolean, nullable=False, default=False)

    #: For diagnostics purpose
    human_readable_description = sa.Column(sa.Text, nullable=False)

    #: Address of the upcoming deployed contract or token contract address interacted with
    contract_address = sa.Column(sa.String(256), nullable=True)

    #: Address of the account, like 0x000000, for benefactor who receives the tokens. Not applicable for contract deployments.
    receiver = sa.Column(sa.String(256), nullable=True)

    #: Raw payload of the transaction to be broadcasted
    unsigned_payload = sa.Column(sa.JSON, nullable=False)

    #: Precalculated transaction id
    txid = sa.Column(sa.String(256), nullable=True)

    #: Value transferred in Ethereum transaction
    # value = sa.Column(sa.Numeric(60, 20), nullable=False, default=0)

    #: How much gwei we paid for this in Ethereum network
    # gas_price = sa.Column(sa.Numeric(60, 20), nullable=False, default=0)

    #: Wat was the gas limit in Etheruem network
    # gas_limit = sa.Column(sa.Numeric(60, 20), nullable=False, default=0)

    #: When we attempted this transaction was broadcasted to the network
    broadcasted_at = sa.Column(UTCDateTime, default=None)

    #: When was the last attempt to rebroadcast this transaction
    #: TODO: Not in use yet.
    rebroadcasted_at = sa.Column(UTCDateTime, default=None)

    #: When did we poll and received that the transaction was included in a block
    result_fetched_at = sa.Column(UTCDateTime, default=None)

    #: What was the resulting block where this transaction was included
    result_block_num = sa.Column(sa.Integer, default=None)

    #: Did the transaction success or fail
    result_transaction_success = sa.Column(sa.Boolean, default=None)

    #: Human readable failure reason
    result_transaction_reason = sa.Column(sa.String(256), default=None)

    #: When a contract deployment was verified at EtherScan
    verified_at = sa.Column(UTCDateTime, default=None)

    #: Misc. transaction data - like ABI with source code information for verification, verification info
    other_data = sa.Column(sa.JSON, nullable=False, default=dict)

    @property
    def gas_limit(self):
        return self.unsigned_payload["gas"]

    @property
    def gas_price(self):
        return self.unsigned_payload["gasPrice"]

    def get_status(self) -> str:
        """Machine/human readable status."""
        if not self.broadcasted_at:
            return "waiting"
        elif self.broadcasted_at and not self.result_fetched_at:
            return "broadcasted"
        elif self.result_fetched_at:

            if self.verified_at:
                return "verified"
            elif self.result_block_num:
                if self.result_transaction_success:
                    return "success"
                else:
                    return "failed"
            else:
                return "mining"
        else:
            raise RuntimeError("State error")

    def get_to(self) -> str:
        return self.receiver or self.contract_address

    def get_from(self) -> str:
        return self.broadcast_account.address

    @property
    def abi(self) -> Optional["str"]:
        """Source code used for a deployment transaction for contract verification."""
        return self.other_data["abi"]["source"]

    @abi.setter
    def abi(self, val):
        self.other_data["abi"] = val
        flag_modified(self, "other_data")

    @property
    def verification_info(self) -> Optional["str"]:
        """EtherScan reply for contract verification."""
        return self.other_data["verification_info"]

    @verification_info.setter
    def verification_info(self, val):
        assert type(val) == dict
        self.other_data["verification_info"] = val
        flag_modified(self, "other_data")

    @property
    def flattened_source_code(self) -> Optional["str"]:
        """Source code used for a deployment transaction for contract verification."""
        return self.other_data["abi"]["source"]

    @property
    def compiler_version(self) -> Optional["str"]:
        """Compiled version used for verification."""
        return self.other_data["abi"]["metadata"]["compiler"]["version"]

    @property
    def contract_name(self) -> Optional["str"]:
        """Compiled version used for verification."""
        abi = self.other_data["abi"]
        return abi["name"]

    @property
    def constructor_arguments(self) -> Optional[str]:
        """The contract payload used for the constructor.

        Needed for source code verification.
        """
        return self.other_data["constructor_arguments"]

    @constructor_arguments.setter
    def constructor_arguments(self, val):
        assert val.startswith("0x")
        self.other_data["constructor_arguments"] = val
        flag_modified(self, "other_data")

    def is_token_contract_deployment(self) -> bool:
        """Is this transaction to deploy a token contract.

        To separate for deploying transfer agents, etc. other smart contracts.
        """
        return self.contract_name in ("SecurityToken",)

